@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2012, 2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -16,7 +17,8 @@ from wimpiggy.lowlevel import (
            remove_event_receiver,           #@UnresolvedImport
            get_parent,                      #@UnresolvedImport
            addXSelectInput, const,          #@UnresolvedImport
-           geometry_with_border             #@UnresolvedImport
+           geometry_with_border,            #@UnresolvedImport
+           get_xwindow                      #@UnresolvedImport
            )
 
 from wimpiggy.log import Logger
@@ -43,53 +45,66 @@ class CompositeHelper(AutoPropGObjectMixin, gobject.GObject):
     # This may raise XError.
     def __init__(self, window, already_composited):
         super(CompositeHelper, self).__init__()
+        log("CompositeHelper.__init__(%s,%s)", window, already_composited)
         self._window = window
         self._already_composited = already_composited
-        def setup():
-            if not self._already_composited:
-                xcomposite_redirect_window(window)
-            (_, _, _, _, self._border_width) = geometry_with_border(window)
-        trap.call(setup)
         self._listening_to = None
-        self.invalidate_pixmap()
-        self._damage_handle = xdamage_start(window)
+        self._damage_handle = None
 
+    def setup(self):
+        if not self._already_composited:
+            xcomposite_redirect_window(self._window)
+        _, _, _, _, self._border_width = geometry_with_border(self._window)
+        self.invalidate_pixmap()
+        self._damage_handle = xdamage_start(self._window)
         add_event_receiver(self._window, self)
 
     def destroy(self):
         if self._window is None:
             log.warn("composite window %s already destroyed!", self)
             return
-        if not self._already_composited:
-            trap.swallow(xcomposite_unredirect_window, self._window)
-        trap.swallow(xdamage_stop, self._window, self._damage_handle)
-        self._damage_handle = None
-        self._contents_handle = None
-        self.invalidate_pixmap()
-        remove_event_receiver(self._window, self)
+        #clear the reference to the window early:
+        win = self._window
+        #Note: invalidate_pixmap()/_cleanup_listening() use self._window, but won't care if it's None
         self._window = None
+        remove_event_receiver(win, self)
+        self.invalidate_pixmap()
+        if not self._already_composited:
+            trap.swallow_synced(xcomposite_unredirect_window, win)
+        if self._damage_handle:
+            trap.swallow_synced(xdamage_stop, win, self._damage_handle)
+            self._damage_handle = None
+        #note: this should be redundant since we cleared the
+        #reference to self._window and shortcut out in do_get_property_contents_handle
+        #but it's cheap anyway
+        self.invalidate_pixmap()
 
     def acknowledge_changes(self):
         if self._damage_handle is not None and self._window is not None:
-            trap.swallow(xdamage_acknowledge, self._window, self._damage_handle)
+            #"Synchronously modifies the regions..." so unsynced?
+            if not trap.swallow_synced(xdamage_acknowledge, self._window, self._damage_handle):
+                self.invalidate_pixmap()
 
     def invalidate_pixmap(self):
-        log("invalidating named pixmap", type="pixmap")
+        log("invalidating named pixmap")
         if self._listening_to is not None:
             self._cleanup_listening(self._listening_to)
             self._listening_to = None
         self._contents_handle = None
 
     def _cleanup_listening(self, listening):
-        # Don't want to stop listening to self._window!:
-        if self._listening_to:
-            assert self._window not in self._listening_to
-            for w in self._listening_to:
+        if listening:
+            # Don't want to stop listening to self._window!:
+            assert self._window is None or self._window not in listening
+            for w in listening:
                 remove_event_receiver(w, self)
 
     def do_get_property_contents_handle(self, name):
+        if self._window is None:
+            #shortcut out
+            return  None
         if self._contents_handle is None:
-            log("refreshing named pixmap", type="pixmap")
+            log("refreshing named pixmap")
             assert self._listening_to is None
             def set_pixmap():
                 # The tricky part here is that the pixmap returned by
@@ -107,6 +122,7 @@ class CompositeHelper(AutoPropGObjectMixin, gobject.GObject):
                 #   4) call NameWindowPixmap
                 # we are safe.  (I think.)
                 listening = []
+                e = None
                 try:
                     win = get_parent(self._window)
                     while win is not None and win.get_parent() is not None:
@@ -118,19 +134,18 @@ class CompositeHelper(AutoPropGObjectMixin, gobject.GObject):
                         # clobbering.  They are our friends!  X is driving me
                         # slowly mad.
                         addXSelectInput(win, const["StructureNotifyMask"])
-                        add_event_receiver(win, self)
+                        add_event_receiver(win, self, max_receivers=-1)
                         listening.append(win)
                         win = get_parent(win)
                     handle = xcomposite_name_window_pixmap(self._window)
-                except:
+                except Exception, e:
                     try:
                         self._cleanup_listening(listening)
                     except:
                         pass
                     raise
                 if handle is None:
-                    log.warn("failed to name a window pixmap (expect an X error soon)",
-                        type="pixmap")
+                    log("failed to name a window pixmap for %s: %s", get_xwindow(self._window), e)
                     self._cleanup_listening(listening)
                 else:
                     self._contents_handle = handle
@@ -138,7 +153,7 @@ class CompositeHelper(AutoPropGObjectMixin, gobject.GObject):
                     # NameWindowPixmap has succeeded, to maintain our
                     # invariant:
                     self._listening_to = listening
-            trap.swallow(set_pixmap)
+            trap.swallow_synced(set_pixmap)
         return self._contents_handle
 
     def do_get_property_contents(self, name):
