@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2012, 2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -105,7 +106,26 @@ class HotkeyManager(gobject.GObject):
 gobject.type_register(HotkeyManager)
 
 
-def grok_modifier_map(display_source):
+DEFAULT_MODIFIER_MEANINGS = {
+        "Shift_L"   : "shift",
+        "Shift_R"   : "shift",
+        "Caps_Lock" : "lock",
+        "Control_L" : "control",
+        "Control_R" : "control",
+        "Alt_L"     : "mod1",
+        "Alt_R"     : "mod1",
+        "Num_Lock"  : "mod2",
+        "Meta_L"    : "mod3",
+        "Meta_R"    : "mod3",
+        "Super_L"   : "mod4",
+        "Super_R"   : "mod4",
+        "Hyper_L"   : "mod4",
+        "Hyper_R"   : "mod4",
+        "ISO_Level3_Shift"  : "mod5",
+        "Mode_switch"       : "mod5",
+        }
+
+def grok_modifier_map(display_source, meanings):
     """Return an dict mapping modifier names to corresponding X modifier
     bitmasks."""
     modifier_map = {
@@ -124,18 +144,8 @@ def grok_modifier_map(display_source):
         "hyper": 0,
         "alt": 0,
         }
-    meanings = {
-        "Scroll_Lock": "scroll",
-        "Num_Lock": "num",
-        "Meta_L": "meta",
-        "Meta_R": "meta",
-        "Super_L": "super",
-        "Super_R": "super",
-        "Hyper_L": "hyper",
-        "Hyper_R": "hyper",
-        "Alt_L": "alt",
-        "Alt_R": "alt",
-        }
+    if not meanings:
+        meanings = DEFAULT_MODIFIER_MEANINGS
 
     disp = get_display_for(display_source)
     (max_keypermod, keycodes) = get_modifier_map(disp)
@@ -151,11 +161,13 @@ def grok_modifier_map(display_source):
                     continue
                 for (keyval, _, _, _) in entries:
                     keyval_name = gtk.gdk.keyval_name(keyval)
-                    if keyval_name in meanings:
-                        modifier_map[meanings[keyval_name]] |= (1 << i)
+                    modifier = meanings.get(keyval_name)
+                    if modifier:
+                        modifier_map[modifier] |= (1 << i)
     modifier_map["nuisance"] = (modifier_map["lock"]
                                 | modifier_map["scroll"]
                                 | modifier_map["num"])
+    log.debug("grok_modifier_map(%s,%s)=%s" % (display_source, meanings, modifier_map))
     return modifier_map
 
 def parse_key(name, keymap, modifier_map):
