@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2012, 2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -23,6 +24,7 @@ from wimpiggy.lowlevel import (
                 PropertyError,              #@UnresolvedImport
                 get_xatom, get_pyatom,      #@UnresolvedImport
                 get_xwindow, get_pywindow,  #@UnresolvedImport
+                get_xvisual,                #@UnresolvedImport
                 const,                      #@UnresolvedImport
                 premultiply_argb_in_place   #@UnresolvedImport
                )
@@ -63,8 +65,6 @@ class WMSizeHints(object):
          max_aspect_num, max_aspect_denom,
          base_width, base_height,
          win_gravity) = struct.unpack("=" + "I" * 18, data) #@UnusedVariable
-        #print(repr(data))
-        #print(struct.unpack("@" + "i" * 18, data))
         # We only extract the pieces we care about:
         if flags & const["PMaxSize"]:
             self.max_size = (max_width, max_height)
@@ -90,6 +90,10 @@ class WMSizeHints(object):
         else:
             self.min_aspect, self.max_aspect = (None, None)
             self.min_aspect_ratio, self.max_aspect_ratio = (None, None)
+        if flags & const["PWinGravity"]:
+            self.win_gravity = win_gravity
+        else:
+            self.win_gravity = -1
 
 class WMHints(object):
     def __init__(self, disp, data):
@@ -193,22 +197,6 @@ def _get_multiple(disp, d):
     return _get_atom(disp, d)
 
 
-def set_xsettings_format(use_tuple=True):
-    log("set_xsettings_format(%s)", use_tuple)
-    global _prop_types
-    if use_tuple:
-        _prop_types["xsettings-settings"] = (tuple, "_XSETTINGS_SETTINGS", 8,
-                           set_settings,
-                           get_settings,
-                           None)
-    else:
-        #for old clients that rely on the old raw string format:
-        _prop_types["xsettings-settings"] = (str, "_XSETTINGS_SETTINGS", 8,
-                           lambda disp, c: c,
-                           lambda disp, d: d,
-                           None)
-
-
 _prop_types = {
     # Python type, X type Atom, formatbits, serializer, deserializer, list
     # terminator
@@ -231,6 +219,10 @@ _prop_types = {
             lambda disp, c: struct.pack("=I", c),
             lambda disp, d: struct.unpack("=I", d)[0],
             ""),
+    "visual": (gtk.gdk.Visual, "XA_VISUALID", 32,
+               lambda disp, c: struct.pack("=I", get_xvisual(c)),
+               unsupported,
+               ""),
     "window": (gtk.gdk.Window, "WINDOW", 32,
                lambda disp, c: struct.pack("=I", get_xwindow(c)),
                lambda disp, d: get_pywindow(disp, struct.unpack("=I", d)[0]),
@@ -289,7 +281,7 @@ def _prop_encode_list(disp, etype, value):
 
 
 def prop_set(target, key, etype, value):
-    trap.call_unsynced(XChangeProperty, target, key,
+    trap.call_synced(XChangeProperty, target, key,
                        _prop_encode(target, etype, value))
 
 def _prop_decode(disp, etype, data):
@@ -319,22 +311,26 @@ def _prop_decode_list(disp, etype, data):
     return [x for x in props if x is not None]
 
 # May return None.
-def prop_get(target, key, etype, ignore_errors=False):
+def prop_get(target, key, etype, ignore_errors=False, raise_xerrors=False):
     if isinstance(etype, list):
         scalar_type = etype[0]
     else:
         scalar_type = etype
     (_, atom, _, _, _, _) = _prop_types[scalar_type]
     try:
-        #print(atom)
         data = trap.call_synced(XGetWindowProperty, target, key, atom)
-        #print(atom, repr(data[:100]))
     except NoSuchProperty:
         log.debug("Missing property %s (%s)", key, etype)
         return None
-    except (XError, PropertyError):
+    except XError:
+        if raise_xerrors:
+            raise
+        log.info("Missing window %s or wrong property type %s (%s)", target, key, etype)
+        traceback.print_exc()
+        return None
+    except PropertyError:
         if not ignore_errors:
-            log.info("Missing window or missing property or wrong property type %s (%s)", key, etype)
+            log.info("Missing property or wrong property type %s (%s)", key, etype)
             traceback.print_exc()
         return None
     try:
