@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -37,7 +38,8 @@ from wimpiggy.lowlevel import (
                XKillClient,                                 #@UnresolvedImport
                sendConfigureNotify,                         #@UnresolvedImport
                configureAndNotify,                          #@UnresolvedImport
-               substructureRedirect                         #@UnresolvedImport
+               substructureRedirect,                        #@UnresolvedImport
+               get_xwindow,                                 #@UnresolvedImport
                )
 from wimpiggy.lowlevel.send_wm import (
                 send_wm_take_focus,                         #@UnresolvedImport
@@ -224,6 +226,9 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         "transient-for": (gobject.TYPE_PYOBJECT,
                           "Transient for (or None)", "",
                           gobject.PARAM_READABLE),
+        "modal": (gobject.TYPE_PYOBJECT,
+                          "Modal (boolean)", "",
+                          gobject.PARAM_READABLE),
         "window-type": (gobject.TYPE_PYOBJECT,
                         "Window type",
                         "NB, most preferred comes first, then fallbacks",
@@ -244,30 +249,62 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
         }
 
     def __init__(self, client_window):
+        log("new window %s - %s", hex(client_window.xid), get_xwindow(client_window))
         super(BaseWindowModel, self).__init__()
-
-        log("new window %s", hex(client_window.xid))
-
-        self._managed = False
-        self._composite = None
         self.client_window = client_window
+        self._managed = False
+        self._setup_done = False
+        self._geometry = None
+        self._damage_forward_handle = None
         self._internal_set_property("client-window", client_window)
-        add_event_receiver(client_window, self)
+        self._composite = CompositeHelper(self.client_window, False)
 
-        def setup():
-            # Keith Packard says that composite state is undefined following a
-            # reparent, so I'm not sure doing this here in the superclass,
-            # before we reparent, actually works... let's wait and see.
-            self._composite = CompositeHelper(self.client_window, False)
-            h = self._composite.connect("contents-changed", self._forward_contents_changed)
-            self._composite.connect("wimpiggy-configure-event", self.composite_configure_event)
-            self._damage_forward_handle = h
-            self._geometry = geometry_with_border(self.client_window)
-            self._managed = True
+    def call_setup(self):
+        log("call_setup() adding event receiver")
         try:
-            trap.call(setup)
+            self._geometry = trap.call_unsynced(geometry_with_border, self.client_window)
         except XError, e:
             raise Unmanageable(e)
+        add_event_receiver(self.client_window, self)
+        # Keith Packard says that composite state is undefined following a
+        # reparent, so I'm not sure doing this here in the superclass,
+        # before we reparent, actually works... let's wait and see.
+        try:
+            trap.call_synced(self._composite.setup)
+        except XError, e:
+            remove_event_receiver(self.client_window, self)
+            log("window %s does not support compositing: %s", get_xwindow(self.client_window), e)
+            trap.swallow_synced(self._composite.destroy)
+            self._composite = None
+            raise Unmanageable(e)
+        #compositing is now enabled, from now on we need to call setup_failed to clean things up
+        try:
+            trap.call_synced(self.setup)
+        except XError, e:
+            try:
+                trap.call_synced(self.setup_failed, e)
+            except Exception, ex:
+                log.error("error in cleanup handler: %s", ex)
+            raise Unmanageable(e)
+        self._managed = True
+        self._setup_done = True
+        log("call_setup() ended")
+
+    def setup_failed(self, e):
+        log("cannot manage %s: %s", get_xwindow(self.client_window), e)
+        self.do_unmanaged(False)
+
+    def setup(self):
+        h = self._composite.connect("contents-changed", self._forward_contents_changed)
+        self._composite.connect("wimpiggy-configure-event", self.composite_configure_event)
+        self._damage_forward_handle = h
+
+    def prop_get(self, key, ptype, ignore_errors=False, raise_xerrors=False):
+        # Utility wrapper for prop_get on the client_window
+        # also allows us to ignore property errors during setup_client
+        if not self._setup_done:
+            ignore_errors = True
+        return prop_get(self.client_window, key, ptype, ignore_errors=ignore_errors, raise_xerrors=raise_xerrors)
 
     def is_managed(self):
         return self._managed
@@ -287,7 +324,7 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
     def do_wimpiggy_configure_event(self, event):
         self._geometry = (event.x, event.y, event.width, event.height,
                           event.border_width)
-        self.notify("geometry")
+        log.info("WindowModel.do_wimpiggy_configure_event(%s)", event)
 
     def composite_configure_event(self, composite_window, event):
         log("BaseWindowModel.composite_configure_event(%s,%s)", composite_window, event)
@@ -300,22 +337,29 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
 
     def unmanage(self, exiting=False):
         if self._managed:
-            self._managed = False
             self.emit("unmanaged", exiting)
 
     def do_unmanaged(self, wm_exiting):
+        if not self._managed:
+            return
+        self._managed = False
+        log("do_unmanaged(%s) damage_forward_handle=%s, composite=%s", wm_exiting, self._damage_forward_handle, self._composite)
         remove_event_receiver(self.client_window, self)
         if self._composite:
-            self._composite.disconnect(self._damage_forward_handle)
+            if self._damage_forward_handle:
+                self._composite.disconnect(self._damage_forward_handle)
+                self._damage_forward_handle = None
             self._composite.destroy()
             self._composite = None
 
     def _read_initial_properties(self):
-        transient_for = self.prop_get("WM_TRANSIENT_FOR", "window")
+        def pget(key, ptype):
+            return self.prop_get(key, ptype, raise_xerrors=True)
+        transient_for = pget("WM_TRANSIENT_FOR", "window")
         # May be None
         self._internal_set_property("transient-for", transient_for)
 
-        window_types = self.prop_get("_NET_WM_WINDOW_TYPE", ["atom"])
+        window_types = pget("_NET_WM_WINDOW_TYPE", ["atom"])
         if not window_types:
             window_type = self._guess_window_type(transient_for)
             window_types = [gtk.gdk.atom_intern(window_type)]
@@ -330,6 +374,8 @@ class BaseWindowModel(AutoPropGObjectMixin, gobject.GObject):
             return "_NET_WM_TYPE_DIALOG"
         return "_NET_WM_WINDOW_TYPE_NORMAL"
 
+    def is_tray(self):
+        return False
 
 gobject.type_register(BaseWindowModel)
 
@@ -344,22 +390,24 @@ class OverrideRedirectWindowModel(BaseWindowModel):
 
     def __init__(self, client_window):
         BaseWindowModel.__init__(self, client_window)
-        def setup():
-            self.client_window.set_events(self.client_window.get_events()
-                                          | gtk.gdk.STRUCTURE_MASK)
-            # So now if the window becomes unmapped in the future then we will
-            # notice... but it might be unmapped already, and any event
-            # already generated, and our request for that event is too late!
-            # So double check now, *after* putting in our request:
-            if not is_mapped(self.client_window):
-                raise Unmanageable("window already unmapped")
 
-            self._managed = True
-            self._read_initial_properties()
-        try:
-            trap.call(setup)
-        except XError, e:
-            raise Unmanageable(e)
+    def call_setup(self):
+        self._read_initial_properties()
+        BaseWindowModel.call_setup(self)
+
+    def setup(self):
+        BaseWindowModel.setup(self)
+        self.client_window.set_events(self.client_window.get_events()
+                                      | gtk.gdk.STRUCTURE_MASK)
+        # So now if the window becomes unmapped in the future then we will
+        # notice... but it might be unmapped already, and any event
+        # already generated, and our request for that event is too late!
+        # So double check now, *after* putting in our request:
+        if not is_mapped(self.client_window):
+            raise Unmanageable("window already unmapped")
+        ch = self._composite.get_property("contents-handle")
+        if ch is None:
+            raise Unmanageable("failed to get damage handle")
 
     def _guess_window_type(self, transient_for):
         return "_NET_WM_WINDOW_TYPE_NORMAL"
@@ -367,20 +415,39 @@ class OverrideRedirectWindowModel(BaseWindowModel):
     def do_wimpiggy_unmap_event(self, event):
         self.unmanage()
 
-    def prop_get(self, key, ptype):
-        return prop_get(self.client_window, key, ptype, ignore_errors=False)
-
     def get_dimensions(self):
-        _, _, ww, wh = self.get_property("geometry")
+        ww, wh = self._geometry[2:4]
         return ww, wh
 
     def is_OR(self):
         return  True
 
-
 gobject.type_register(OverrideRedirectWindowModel)
 
 
+class SystemTrayWindowModel(OverrideRedirectWindowModel):
+
+    def __init__(self, client_window):
+        OverrideRedirectWindowModel.__init__(self, client_window)
+
+    def is_tray(self):
+        return  True
+
+    def _read_initial_properties(self):
+        pass
+
+    def composite_configure_event(self, composite_window, event):
+        BaseWindowModel.composite_configure_event(self, composite_window, event)
+        log("SystemTrayWindowModel.composite_configure_event(%s, %s) client window geometry=%s", composite_window, event, self.client_window.get_geometry())
+
+    def move_resize(self, x, y, width, height):
+        #Used by clients to tell us where the tray is located on screen
+        log("SystemTrayWindowModel.move_resize(%s, %s, %s, %s)", x, y, width, height)
+        self.client_window.move_resize(x, y, width, height)
+        border = self._geometry[4]
+        self._geometry = (x, y, width, height, border)
+
+
 class WindowModel(BaseWindowModel):
     """This represents a managed client window.  It allows one to produce
     widgets that view that client window in various ways."""
@@ -493,10 +560,21 @@ class WindowModel(BaseWindowModel):
 
         BaseWindowModel.__init__(self, client_window)
         self.parking_window = parking_window
-        self._setup_done = False
+        self.corral_window = None
+        self.client_window_saved_events = self.client_window.get_events()
+        self.in_save_set = False
+        self.client_reparented = False
+        self.startup_unmap_serial = None
 
+        # The WM_HINTS input field
+        self._input_field = True
         self.connect("notify::iconic", self._handle_iconic_update)
 
+        self.call_setup()
+
+    def setup(self):
+        BaseWindowModel.setup(self)
+
         # We enable PROPERTY_CHANGE_MASK so that we can call
         # x11_get_server_time on this window.
         self.corral_window = gtk.gdk.Window(self.parking_window,
@@ -504,54 +582,55 @@ class WindowModel(BaseWindowModel):
                                             height=100,
                                             window_type=gtk.gdk.WINDOW_CHILD,
                                             wclass=gtk.gdk.INPUT_OUTPUT,
-                                            event_mask=gtk.gdk.PROPERTY_CHANGE_MASK)
+                                            event_mask=gtk.gdk.PROPERTY_CHANGE_MASK,
+                                            title = "CorralWindow-0x%s" % self.client_window.xid)
+        log("setup() corral_window=%s", self.corral_window)
         substructureRedirect(self.corral_window)
         add_event_receiver(self.corral_window, self)
-        log("created corral window 0x%x", self.corral_window.xid)
 
-        # The WM_HINTS input field
-        self._input_field = True
+        # Start listening for important events.
+        self.client_window.set_events(self.client_window_saved_events
+                                      | gtk.gdk.STRUCTURE_MASK
+                                      | gtk.gdk.PROPERTY_CHANGE_MASK)
 
-        def setup_client():
-            # Start listening for important events.
-            self.client_window.set_events(self.client_window.get_events()
-                                          | gtk.gdk.STRUCTURE_MASK
-                                          | gtk.gdk.PROPERTY_CHANGE_MASK)
+        # The child might already be mapped, in case we inherited it from
+        # a previous window manager.  If so, we unmap it now, and save the
+        # serial number of the request -- this way, when we get an
+        # UnmapNotify later, we'll know that it's just from us unmapping
+        # the window, not from the client withdrawing the window.
+        if is_mapped(self.client_window):
+            log("hiding inherited window")
+            self.startup_unmap_serial = unmap_with_serial(self.client_window)
 
-            # The child might already be mapped, in case we inherited it from
-            # a previous window manager.  If so, we unmap it now, and save the
-            # serial number of the request -- this way, when we get an
-            # UnmapNotify later, we'll know that it's just from us unmapping
-            # the window, not from the client withdrawing the window.
-            self.startup_unmap_serial = None
-            if is_mapped(self.client_window):
-                log("hiding inherited window")
-                self.startup_unmap_serial = unmap_with_serial(self.client_window)
+        # Process properties
+        self._read_initial_properties()
+        self._write_initial_properties_and_setup()
 
-            # Process properties
-            self._read_initial_properties()
-            self._write_initial_properties_and_setup()
+        # For now, we never use the Iconic state at all.
+        self._internal_set_property("iconic", False)
 
-            # For now, we never use the Iconic state at all.
-            self._internal_set_property("iconic", False)
+        log("setup() adding to save set")
+        XAddToSaveSet(self.client_window)
+        self.in_save_set = True
+
+        log("setup() reparenting")
+        self.client_window.reparent(self.corral_window, 0, 0)
+        self.client_reparented = True
+
+        log("setup() geometry")
+        w,h = self.client_window.get_geometry()[2:4]
+        hints = self.get_property("size-hints")
+        self._sanitize_size_hints(hints)
+        nw, nh = calc_constrained_size(w, h, hints)[:2]
+        if nw>=32768 or nh>=32768:
+            #we can't handle windows that big!
+            raise Unmanageable("window constrained size is too large: %sx%s (from client geometry: %s,%s with size hints=%s)" % (nw, nh, w, h, hints))
+        log("setup() resizing windows to %sx%s", nw, nh)
+        self.client_window.resize(nw, nh)
+        self.corral_window.resize(nw, nh)
+        self.client_window.show_unraised()
+        self.client_window.get_geometry()
 
-            XAddToSaveSet(self.client_window)
-            self.client_window.reparent(self.corral_window, 0, 0)
-            w,h = self.client_window.get_geometry()[2:4]
-            hints = self.get_property("size-hints")
-            self._sanitize_size_hints(hints)
-            client_size = calc_constrained_size(w, h, hints)[:2]
-            log("setup_client() resizing windows to %s", client_size)
-            self.client_window.resize(*client_size)
-            self.corral_window.resize(*client_size)
-            self.client_window.show_unraised()
-            self.client_window.get_geometry()
-        try:
-            trap.call(setup_client)
-            trap.call(self.client_window.get_geometry)
-        except XError, e:
-            raise Unmanageable(e)
-        self._setup_done = True
 
     def is_OR(self):
         return  False
@@ -559,13 +638,6 @@ class WindowModel(BaseWindowModel):
     def get_dimensions(self):
         return  self.get_property("actual-size")
 
-    def prop_get(self, key, ptype, ignore_errors=False):
-        # Utility wrapper for prop_get on the client_window
-        # also allows us to ignore property errors during setup_client
-        if not self._setup_done:
-            ignore_errors = True
-        return prop_get(self.client_window, key, ptype, ignore_errors=ignore_errors)
-
     def do_wimpiggy_xkb_event(self, event):
         log("WindowModel.do_wimpiggy_xkb_event(%r)" % event)
         if event.type!="bell":
@@ -616,28 +688,39 @@ class WindowModel(BaseWindowModel):
         # the morning.  It makes for simple code:
         self.unmanage()
 
-    def do_unmanaged(self, exiting):
+    SCRUB_PROPERTIES = ["WM_STATE",
+                        "_NET_WM_STATE",
+                        "_NET_FRAME_EXTENTS",
+                        "_NET_WM_ALLOWED_ACTIONS",
+                        ]
+
+    def do_unmanaged(self, wm_exiting):
         log("unmanaging window: %s (%s - %s)", self, self.corral_window, self.client_window)
         self._internal_set_property("owner", None)
-        def unmanageit():
-            self._scrub_withdrawn_window()
-            self.client_window.reparent(gtk.gdk.get_default_root_window(),
-                                        0, 0)
+        if self.corral_window:
+            remove_event_receiver(self.corral_window, self)
+            for prop in WindowModel.SCRUB_PROPERTIES:
+                trap.swallow_synced(XDeleteProperty, self.client_window, prop)
+            if self.client_reparented:
+                self.client_window.reparent(gtk.gdk.get_default_root_window(), 0, 0)
+                self.client_reparented = False
+            self.client_window.set_events(self.client_window_saved_events)
+            #it is now safe to destroy the corral window:
+            self.corral_window.destroy()
+            self.corral_window = None
             # It is important to remove from our save set, even after
             # reparenting, because according to the X spec, windows that are
             # in our save set are always Mapped when we exit, *even if those
             # windows are no longer inferior to any of our windows!* (see
             # section 10. Connection Close).  This causes "ghost windows", see
             # bug #27:
-            XRemoveFromSaveSet(self.client_window)
-            sendConfigureNotify(self.client_window)
-            if exiting:
+            if self.in_save_set:
+                trap.swallow_synced(XRemoveFromSaveSet, self.client_window)
+                self.in_save_set = False
+            trap.swallow_synced(sendConfigureNotify, self.client_window)
+            if wm_exiting:
                 self.client_window.show_unraised()
-        if self.corral_window:
-            trap.swallow(unmanageit)
-            self.corral_window.destroy()
-            self.corral_window = None
-        BaseWindowModel.do_unmanaged(self, exiting)
+        BaseWindowModel.do_unmanaged(self, wm_exiting)
 
     def ownership_election(self):
         candidates = self.emit("ownership-election")
@@ -658,7 +741,11 @@ class WindowModel(BaseWindowModel):
             winner.take_window(self, self.corral_window)
             self._update_client_geometry()
             self.corral_window.show_unraised()
-        trap.swallow(sendConfigureNotify, self.client_window)
+        trap.swallow_synced(sendConfigureNotify, self.client_window)
+
+    def do_wimpiggy_configure_event(self, event):
+        WindowModel.do_wimpiggy_configure_event(self, event)
+        self.notify("geometry")
 
     def maybe_recalculate_geometry_for(self, maybe_owner):
         if maybe_owner and self.get_property("owner") is maybe_owner:
@@ -668,7 +755,6 @@ class WindowModel(BaseWindowModel):
         if size_hints is None:
             return
         for attr in ["min_aspect", "max_aspect"]:
-            #"min_aspect_ratio", "max_aspect_ratio" ??
             v = getattr(size_hints, attr)
             if v is not None:
                 try:
@@ -678,8 +764,8 @@ class WindowModel(BaseWindowModel):
                 if f is None or f>=(2**32-1):
                     log.warn("clearing invalid aspect hint value for %s: %s", attr, v)
                     setattr(size_hints, attr, -1.0)
-        for attr in ["max_size", "min_size", "base_size", "resize_inc"]:
-            #"min_aspect_ratio", "max_aspect_ratio" ??
+        for attr in ["max_size", "min_size", "base_size", "resize_inc",
+                     "min_aspect_ratio", "max_aspect_ratio"]:
             v = getattr(size_hints, attr)
             if v is not None:
                 try:
@@ -711,7 +797,7 @@ class WindowModel(BaseWindowModel):
         x, y = window_position_cb(w, h)
         log("_do_update_client_geometry: position=%s", (x,y))
         self.corral_window.move_resize(x, y, w, h)
-        trap.swallow(configureAndNotify, self.client_window, 0, 0, w, h)
+        trap.swallow_synced(configureAndNotify, self.client_window, 0, 0, w, h)
         self._internal_set_property("actual-size", (w, h))
         self._internal_set_property("user-friendly-size", (wvis, hvis))
 
@@ -729,7 +815,7 @@ class WindowModel(BaseWindowModel):
             return
         try:
             #workaround applications whose windows disappear from underneath us:
-            if trap.call(self.resize_corral_window):
+            if trap.call_synced(self.resize_corral_window):
                 self.emit("geometry")
         except XError, e:
             log.warn("failed to resize corral window: %s", e)
@@ -760,7 +846,7 @@ class WindowModel(BaseWindowModel):
         # Ignore the request, but as per ICCCM 4.1.5, send back a synthetic
         # ConfigureNotify telling the client that nothing has happened.
         log("do_child_configure_request_event(%s)", event)
-        trap.swallow(sendConfigureNotify, event.window)
+        trap.swallow_synced(sendConfigureNotify, event.window)
 
         # Also potentially update our record of what the app has requested:
         (x, y) = self.get_property("requested-position")
@@ -892,6 +978,8 @@ class WindowModel(BaseWindowModel):
     def _read_initial_properties(self):
         # Things that don't change:
         BaseWindowModel._read_initial_properties(self)
+        def pget(key, ptype):
+            return self.prop_get(key, ptype, raise_xerrors=True)
 
         geometry = self.client_window.get_geometry()
         self._internal_set_property("requested-position", (geometry[0], geometry[1]))
@@ -908,24 +996,24 @@ class WindowModel(BaseWindowModel):
             except ValueError:
                 log.warn("Malformed WM_CLASS: %s, ignoring", class_instance)
                 return  False
-        class_instance = self.prop_get("WM_CLASS", "latin1")
+        class_instance = pget("WM_CLASS", "latin1")
         if class_instance:
             if not set_class_instance(class_instance):
-                set_class_instance(self.prop_get("WM_CLASS", "utf8"))
+                set_class_instance(pget("WM_CLASS", "utf8"))
 
-        protocols = self.prop_get("WM_PROTOCOLS", ["atom"])
+        protocols = pget("WM_PROTOCOLS", ["atom"])
         if protocols is None:
             protocols = []
         self._internal_set_property("protocols", protocols)
         self.notify("can-focus")
 
-        pid = self.prop_get("_NET_WM_PID", "u32")
+        pid = pget("_NET_WM_PID", "u32")
         if pid is not None:
             self._internal_set_property("pid", pid)
         else:
             self._internal_set_property("pid", -1)
 
-        client_machine = self.prop_get("WM_CLIENT_MACHINE", "latin1")
+        client_machine = pget("WM_CLIENT_MACHINE", "latin1")
         # May be None
         self._internal_set_property("client-machine", client_machine)
 
@@ -944,11 +1032,13 @@ class WindowModel(BaseWindowModel):
         # initial states are read off from the client, and (2) is accomplished
         # by having WM_HINTS affect _NET_WM_STATE.  But this means that
         # WM_HINTS and _NET_WM_STATE handling become intertangled.
-        net_wm_state = self.prop_get("_NET_WM_STATE", ["atom"])
+        net_wm_state = pget("_NET_WM_STATE", ["atom"])
         if net_wm_state:
             self._internal_set_property("state", ImmutableSet(net_wm_state))
         else:
             self._internal_set_property("state", ImmutableSet())
+        modal = (net_wm_state is not None) and ("_NET_WM_STATE_MODAL" in net_wm_state)
+        self._internal_set_property("modal", modal)
 
         for mutable in ["WM_HINTS", "WM_NORMAL_HINTS",
                         "WM_NAME", "_NET_WM_NAME",
@@ -1011,7 +1101,7 @@ class WindowModel(BaseWindowModel):
 
     def _handle_state_changed(self, *args):
         # Sync changes to "state" property out to X property.
-        prop_set(self.client_window, "_NET_WM_STATE",
+        trap.swallow_synced(prop_set, self.client_window, "_NET_WM_STATE",
                  ["atom"], self.get_property("state"))
 
     def do_set_property(self, pspec, value):
@@ -1036,15 +1126,16 @@ class WindowModel(BaseWindowModel):
 
 
     def _handle_iconic_update(self, *args):
+        def set_state(state):
+            trap.swallow_synced(prop_set, self.client_window, "WM_STATE",
+                             ["u32"],
+                             [state, const["XNone"]])
+
         if self.get_property("iconic"):
-            trap.swallow(prop_set, self.client_window, "WM_STATE",
-                         ["u32"],
-                         [const["IconicState"], const["XNone"]])
+            set_state(const["IconicState"])
             self._state_add("_NET_WM_STATE_HIDDEN")
         else:
-            trap.swallow(prop_set, self.client_window, "WM_STATE",
-                         ["u32"],
-                         [const["NormalState"], const["XNone"]])
+            set_state(const["NormalState"])
             self._state_remove("_NET_WM_STATE_HIDDEN")
 
     def _write_initial_properties_and_setup(self):
@@ -1058,16 +1149,6 @@ class WindowModel(BaseWindowModel):
         # Flush things:
         self._handle_state_changed()
 
-    def _scrub_withdrawn_window(self):
-        remove = ["WM_STATE",
-                  "_NET_WM_STATE",
-                  "_NET_FRAME_EXTENTS",
-                  "_NET_WM_ALLOWED_ACTIONS",
-                  ]
-        def delete_properties():
-            for prop in remove:
-                XDeleteProperty(self.client_window, prop)
-        trap.swallow(delete_properties)
 
     ################################
     # Focus handling:
@@ -1076,8 +1157,10 @@ class WindowModel(BaseWindowModel):
     def give_client_focus(self):
         """The focus manager has decided that our client should recieve X
         focus.  See world_window.py for details."""
-        if self.corral_window is None:
-            return  True
+        if self.corral_window:
+            trap.swallow_synced(self.do_give_client_focus)
+
+    def do_give_client_focus(self):
         log("Giving focus to client")
         # Have to fetch the time, not just use CurrentTime, both because ICCCM
         # says that WM_TAKE_FOCUS must use a real time and because there are
@@ -1098,18 +1181,12 @@ class WindowModel(BaseWindowModel):
         # XSetInputFocus well, while Qt apps ignore (!!!) WM_TAKE_FOCUS
         # (unless they have a modal window), and just expect to get focus from
         # the WM's XSetInputFocus.
-        success = True
         if self._input_field:
             log("... using XSetInputFocus")
-            if not trap.swallow(XSetInputFocus, self.client_window, now):
-                log("XSetInputFocus failed...")
-                success = False
+            XSetInputFocus(self.client_window, now)
         if "WM_TAKE_FOCUS" in self.get_property("protocols"):
             log("... using WM_TAKE_FOCUS")
-            if not trap.swallow(send_wm_take_focus, self.client_window, now):
-                log("WM_TAKE_FOCUS failed...")
-                success = False
-        return success
+            send_wm_take_focus(self.client_window, now)
 
     ################################
     # Killing clients:
@@ -1117,8 +1194,9 @@ class WindowModel(BaseWindowModel):
 
     def request_close(self):
         if "WM_DELETE_WINDOW" in self.get_property("protocols"):
-            trap.swallow(send_wm_delete_window, self.client_window)
+            trap.swallow_synced(send_wm_delete_window, self.client_window)
         else:
+            log.warn("window does not support WM_DELETE_WINDOW... using force_quit()")
             # You don't wanna play ball?  Then no more Mr. Nice Guy!
             self.force_quit()
 
@@ -1127,11 +1205,14 @@ class WindowModel(BaseWindowModel):
         machine = self.get_property("client-machine")
         localhost = gethostname()
         if pid > 0 and machine is not None and machine == localhost:
-            try:
-                os.kill(pid, 9)
-            except OSError:
-                log.warn("failed to kill() client with pid %s", pid)
-        trap.swallow(XKillClient, self.client_window)
+            if pid==os.getpid():
+                log.warn("force_quit() refusing to kill ourselves!")
+            else:
+                try:
+                    os.kill(pid, 9)
+                except OSError:
+                    log.warn("failed to kill() client with pid %s", pid)
+        trap.swallow_synced(XKillClient, self.client_window)
 
 gobject.type_register(WindowModel)
 
@@ -1362,7 +1443,8 @@ class WindowView(gtk.Widget):
                                      height=self.allocation.height,
                                      window_type=gtk.gdk.WINDOW_CHILD,
                                      wclass=gtk.gdk.INPUT_OUTPUT,
-                                     event_mask=0)
+                                     event_mask=0,
+                                     title="WindowModel-Window")
         self.window.set_user_data(self)
 
         # Give it a nice theme-defined background
@@ -1375,7 +1457,8 @@ class WindowView(gtk.Widget):
                                             height=self.allocation.height,
                                             window_type=gtk.gdk.WINDOW_CHILD,
                                             wclass=gtk.gdk.INPUT_OUTPUT,
-                                            event_mask=gtk.gdk.EXPOSURE_MASK)
+                                            event_mask=gtk.gdk.EXPOSURE_MASK,
+                                            title="WindowModel-ImageWindow")
         self._image_window.input_shape_combine_region(gtk.gdk.Region(),
                                                       0, 0)
         self._image_window.set_user_data(self)
