@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2008, 2009 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2012, 2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -50,8 +51,9 @@ class Wm(gobject.GObject):
         "_NET_DESKTOP_GEOMETRY",
         "_NET_NUMBER_OF_DESKTOPS",
         "_NET_DESKTOP_NAMES",
-        #FIXME: "_NET_WORKAREA",
+        "_NET_WORKAREA",
         "_NET_ACTIVE_WINDOW",
+        "_NET_CURRENT_DESKTOP",
 
         "WM_NAME", "_NET_WM_NAME",
         "WM_ICON_NAME", "_NET_WM_ICON_NAME",
@@ -102,8 +104,8 @@ class Wm(gobject.GObject):
 
         "_NET_WM_STATE",
         "_NET_WM_STATE_DEMANDS_ATTENTION",
+        "_NET_WM_STATE_MODAL",
         # More states to support:
-        # _NET_WM_STATE_MODAL,
         # _NET_WM_STATE_STICKY,
         # _NET_WM_STATE_MAXIMIZED_VERT,
         # _NET_WM_STATE_MAXIMIZED_HORZ,
@@ -118,7 +120,6 @@ class Wm(gobject.GObject):
         # Not at all yet:
         #"_NET_REQUEST_FRAME_EXTENTS",
         #"_NET_CLOSE_WINDOW",
-        #"_NET_CURRENT_DESKTOP",
         #"_NET_RESTACK_WINDOW",
         #"_NET_WM_DESKTOP",
         ]
@@ -187,10 +188,14 @@ class Wm(gobject.GObject):
 
         # Set up the necessary EWMH properties on the root window.
         self._setup_ewmh_window()
-        prop_set(self._root, "_NET_SUPPORTED",
-                 ["atom"], self._NET_SUPPORTED)
-        prop_set(self._root, "_NET_DESKTOP_VIEWPORT",
-                 ["u32"], [0, 0])
+        # Start with just one desktop:
+        self.do_desktop_list_changed([u"Main"])
+        self.set_current_desktop(0)
+        # Start with the full display as workarea:
+        root_w, root_h = gtk.gdk.get_default_root_window().get_size()
+        self.set_workarea(0, 0, root_w, root_h)
+        self.root_set("_NET_DESKTOP_VIEWPORT", ["u32"], [0, 0])
+        self.root_set("_NET_SUPPORTED", ["atom"], self._NET_SUPPORTED)
 
         # Load up our full-screen widget
         self._world_window = WorldWindow()
@@ -222,6 +227,12 @@ class Wm(gobject.GObject):
         # Tray's need to provide info for _NET_ACTIVE_WINDOW and _NET_WORKAREA
         # (and notifications for both)
 
+    def root_set(self, *args):
+        prop_set(self._root, *args)
+
+    def set_workarea(self, x, y, width, height):
+        self.root_set("_NET_WORKAREA", ["u32"], [x, y, width, height])
+
     def enableCursors(self, on):
         log("enableCursors(%s)" % on)
         selectCursorChange(self._root, on)
@@ -247,9 +258,13 @@ class Wm(gobject.GObject):
     # This is in some sense the key entry point to the entire WM program.  We
     # have detected a new client window, and start managing it:
     def _manage_client(self, gdkwindow):
-        if gdkwindow in self._windows:
-            log.error("window %s is already managed!", gdkwindow)
-            return
+        try:
+            if gdkwindow not in self._windows:
+                trap.call_synced(self.do_manage_client, gdkwindow)
+        except Exception, e:
+            log("failed to manage client %s: %s", gdkwindow, e)
+
+    def do_manage_client(self, gdkwindow):
         try:
             win = WindowModel(self._root, gdkwindow)
         except Unmanageable:
@@ -283,10 +298,10 @@ class Wm(gobject.GObject):
         # in a moment we'll get a signal telling us about the window that
         # doesn't exist anymore, will remove it from the list, and then call
         # _update_window_list again.
-        trap.swallow(prop_set, self._root, "_NET_CLIENT_LIST",
+        trap.swallow_synced(self.root_set, "_NET_CLIENT_LIST",
                      ["window"], self._windows_in_order)
         # This is a lie, but we don't maintain a stacking order, so...
-        trap.swallow(prop_set, self._root, "_NET_CLIENT_LIST_STACKING",
+        trap.swallow_synced(self.root_set, "_NET_CLIENT_LIST_STACKING",
                      ["window"], self._windows_in_order)
 
     def do_wimpiggy_client_message_event(self, event):
@@ -327,7 +342,7 @@ class Wm(gobject.GObject):
         if event.window in self._windows:
             return
         log("Reconfigure on withdrawn window")
-        trap.swallow(configureAndNotify,
+        trap.swallow_synced(configureAndNotify,
                      event.window, event.x, event.y,
                      event.width, event.height,
                      event.value_mask)
@@ -344,8 +359,11 @@ class Wm(gobject.GObject):
         printFocus(self._display)
 
     def do_desktop_list_changed(self, desktops):
-        prop_set(self._root, "_NET_NUMBER_OF_DESKTOPS", "u32", len(desktops))
-        prop_set(self._root, "_NET_DESKTOP_NAMES", ["utf8"], desktops)
+        self.root_set("_NET_NUMBER_OF_DESKTOPS", "u32", len(desktops))
+        self.root_set("_NET_DESKTOP_NAMES", ["utf8"], desktops)
+
+    def set_current_desktop(self, index):
+        self.root_set("_NET_CURRENT_DESKTOP", "u32", index)
 
     def _setup_ewmh_window(self):
         # Set up a 1x1 invisible unmapped window, with which to participate in
@@ -366,10 +384,10 @@ class Wm(gobject.GObject):
                                            window_type=gtk.gdk.WINDOW_TOPLEVEL,
                                            event_mask=0, # event mask
                                            wclass=gtk.gdk.INPUT_ONLY,
-                                           title=self._name)
+                                           title="%s-EWMH" % self._name)
         prop_set(self._ewmh_window, "_NET_SUPPORTING_WM_CHECK",
                  "window", self._ewmh_window)
-        prop_set(self._root, "_NET_SUPPORTING_WM_CHECK",
+        self.root_set("_NET_SUPPORTING_WM_CHECK",
                  "window", self._ewmh_window)
 
     # Other global actions:
