@@ -1,21 +1,22 @@
 # This file is part of Parti.
-# Copyright (C) 2010-2012 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2013 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2010 Nathaniel Smith <njs@pobox.com>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
 import sys
-import hashlib
-from wimpiggy.gobject_compat import import_gobject
+import uuid
+import socket
+from wimpiggy.gobject_compat import import_gobject, import_glib
 gobject = import_gobject()
 
 from wimpiggy.util import n_arg_signal
 from wimpiggy.log import Logger
 log = Logger()
 
-from xpra.protocol import Protocol, has_rencode
-from xpra.scripts.main import ENCODINGS
+from xpra.protocol import Protocol, has_rencode, use_rencode
+from xpra.scripts.main import ENCODINGS, ENCRYPTION_CIPHERS
 from xpra.version_util import is_compatible_with, add_version_info
 from xpra.platform import get_machine_id
 
@@ -30,45 +31,9 @@ EXIT_TIMEOUT = 2
 EXIT_PASSWORD_REQUIRED = 3
 EXIT_PASSWORD_FILE_ERROR = 4
 EXIT_INCOMPATIBLE_VERSION = 5
+EXIT_ENCRYPTION = 6
 
-class ClientSource(object):
-    def __init__(self, protocol):
-        self._priority_packets = []
-        self._ordinary_packets = []
-        self._mouse_position = None
-        self._protocol = protocol
-        self._protocol.source = self
-
-    def queue_priority_packet(self, packet):
-        self._priority_packets.append(packet)
-        self._protocol.source_has_more()
-
-    def queue_ordinary_packet(self, packet):
-        self._ordinary_packets.append(packet)
-        self._protocol.source_has_more()
-
-    def queue_positional_packet(self, packet):
-        self.queue_ordinary_packet(packet)
-        self._mouse_position = None
-
-    def queue_mouse_position_packet(self, packet):
-        self._mouse_position = packet
-        self._protocol.source_has_more()
-
-    def next_packet(self):
-        if self._priority_packets:
-            packet = self._priority_packets.pop(0)
-        elif self._ordinary_packets:
-            packet = self._ordinary_packets.pop(0)
-        elif self._mouse_position is not None:
-            packet = self._mouse_position
-            self._mouse_position = None
-        else:
-            packet = None
-        has_more = packet is not None and \
-                (bool(self._priority_packets) or bool(self._ordinary_packets) \
-                 or self._mouse_position is not None)
-        return packet, None, None, has_more
+DEFAULT_TIMEOUT = 20*1000
 
 
 class XpraClientBase(gobject.GObject):
@@ -90,27 +55,41 @@ class XpraClientBase(gobject.GObject):
         self.compression_level = opts.compression_level
         self.password = None
         self.password_file = opts.password_file
+        self.password_sent = False
         self.encoding = opts.encoding
+        self.encryption = opts.encryption
         self.quality = opts.quality
+        self.min_quality = opts.min_quality
+        self.speed = opts.speed
+        self.min_speed = opts.min_speed
+        #protocol stuff:
         self._protocol = None
+        self._priority_packets = []
+        self._ordinary_packets = []
+        self._mouse_position = None
+        #server state and caps:
         self.server_capabilities = {}
         self._remote_version = None
         self._remote_revision = None
+        self.make_uuid()
         self.init_packet_handlers()
 
     def ready(self, conn):
         log.debug("ready(%s)", conn)
-        self._protocol = Protocol(conn, self.process_packet)
+        self._protocol = Protocol(conn, self.process_packet, self.next_packet)
+        self._protocol.large_packets.append("keymap-changed")
+        self._protocol.large_packets.append("server-settings")
         self._protocol.set_compression_level(self.compression_level)
-        ClientSource(self._protocol)
         self._protocol.start()
+        self.have_more = self._protocol.source_has_more
 
     def init_packet_handlers(self):
-        self._packet_handlers = {}
+        self._packet_handlers = {
+            "hello": self._process_hello,
+            }
         self._ui_packet_handlers = {
             "challenge": self._process_challenge,
             "disconnect": self._process_disconnect,
-            "hello": self._process_hello,
             "set_deflate": self._process_set_deflate,
             Protocol.CONNECTION_LOST: self._process_connection_lost,
             Protocol.GIBBERISH: self._process_gibberish,
@@ -119,47 +98,106 @@ class XpraClientBase(gobject.GObject):
     def send_hello(self, challenge_response=None):
         hello = self.make_hello(challenge_response)
         log.debug("send_hello(%s) packet=%s", challenge_response, hello)
-        self.send(["hello", hello])
+        self.send("hello", hello)
 
     def make_hello(self, challenge_response=None):
         capabilities = {}
         add_version_info(capabilities)
         if challenge_response:
+            assert self.password
             capabilities["challenge_response"] = challenge_response
+        if self.encryption:
+            assert self.encryption in ENCRYPTION_CIPHERS
+            capabilities["cipher"] = self.encryption
+            iv = uuid.uuid4().hex[:16]
+            capabilities["cipher.iv"] = iv
+            key_salt = uuid.uuid4().hex
+            capabilities["cipher.key_salt"] = key_salt
+            iterations = 1000
+            capabilities["cipher.key_stretch_iterations"] = iterations
+            self._protocol.set_cipher_in(self.encryption, iv, self.get_password(), key_salt, iterations)
         if self.encoding:
             capabilities["encoding"] = self.encoding
         capabilities["encodings"] = ENCODINGS
-        if self.quality>=0:
+        if self.quality>0:
             capabilities["jpeg"] = self.quality
             capabilities["quality"] = self.quality
+            capabilities["encoding.quality"] = self.quality
+        if self.min_quality>0:
+            capabilities["encoding.min-quality"] = self.min_quality
+        if self.speed>=0:
+            capabilities["speed"] = self.speed
+            capabilities["encoding.speed"] = self.speed
+        if self.min_speed>=0:
+            capabilities["encoding.min-speed"] = self.min_speed
         capabilities["platform"] = sys.platform
         capabilities["client_type"] = "Python/Gobject"
         capabilities["raw_packets"] = True
         capabilities["chunked_compression"] = True
         capabilities["rencode"] = has_rencode
         capabilities["server-window-resize"] = True
-        u = hashlib.sha512()
-        u.update(str(get_machine_id()))
-        if os.name=="posix":
-            u.update("/")
-            u.update(str(os.getuid()))
-            u.update("/")
-            u.update(str(os.getgid()))
-        capabilities["uuid"] = u.hexdigest()
+        capabilities["hostname"] = socket.gethostname()
+        capabilities["uuid"] = self.uuid
         capabilities["randr_notify"] = False    #only client.py cares about this
         capabilities["windows"] = False         #only client.py cares about this
         return capabilities
 
-    def idle_send(self, packet):
-        gobject.idle_add(self.send, packet)
+    def make_uuid(self):
+        try:
+            import hashlib
+            u = hashlib.sha1()
+        except:
+            #try python2.4 variant:
+            import sha
+            u = sha.new()
+        def uupdate(ustr):
+            u.update(ustr.encode("utf-8"))
+        uupdate(get_machine_id())
+        if os.name=="posix":
+            uupdate(u"/")
+            uupdate(str(os.getuid()))
+            uupdate(u"/")
+            uupdate(str(os.getgid()))
+        self.uuid = u.hexdigest()
 
-    def send(self, packet):
-        if self._protocol and self._protocol.source:
-            self._protocol.source.queue_ordinary_packet(packet)
+    def send(self, *parts):
+        self._ordinary_packets.append(parts)
+        self.have_more()
+
+    def send_now(self, *parts):
+        self._priority_packets.append(parts)
+        self.have_more()
+
+    def send_positional(self, packet):
+        self._ordinary_packets.append(packet)
+        self._mouse_position = None
+        self.have_more()
+
+    def send_mouse_position(self, packet):
+        self._mouse_position = packet
+        self.have_more()
+
+    def have_more(self):
+        #this function is overridden in ready()
+        p = self._protocol
+        if p and p.source:
+            p.source_has_more()
+
+    def next_packet(self):
+        if self._priority_packets:
+            packet = self._priority_packets.pop(0)
+        elif self._ordinary_packets:
+            packet = self._ordinary_packets.pop(0)
+        elif self._mouse_position is not None:
+            packet = self._mouse_position
+            self._mouse_position = None
+        else:
+            packet = None
+        has_more = packet is not None and \
+                (bool(self._priority_packets) or bool(self._ordinary_packets) \
+                 or self._mouse_position is not None)
+        return packet, None, None, has_more
 
-    def send_now(self, packet):
-        if self._protocol and self._protocol.source:
-            self._protocol.source.queue_priority_packet(packet)
 
     def cleanup(self):
         if self._protocol:
@@ -194,12 +232,36 @@ class XpraClientBase(gobject.GObject):
             return
         if not self.password:
             self.load_password()
-            log("password read from file %s is %s", self.password_file, self.password)
-        if self.password:
-            salt = packet[1]
-            import hmac
-            challenge_response = hmac.HMAC(self.password, salt)
-            self.send_hello(challenge_response.hexdigest())
+            assert self.password
+        salt = packet[1]
+        if self.encryption:
+            assert len(packet)>=3, "challenge does not contain encryption details to use for the response"
+            server_cipher = packet[2]
+            self.set_server_encryption(server_cipher)
+        import hmac
+        challenge_response = hmac.HMAC(self.password, salt)
+        password_hash = challenge_response.hexdigest()
+        self.password_sent = True
+        self.send_hello(password_hash)
+
+    def set_server_encryption(self, props):
+        cipher = props.get("cipher")
+        cipher_iv = props.get("cipher.iv")
+        key_salt = props.get("cipher.key_salt")
+        iterations = props.get("cipher.key_stretch_iterations")
+        if not cipher or not cipher_iv:
+            self.warn_and_quit(EXIT_ENCRYPTION, "the server does not use or support encryption/password, cannot continue with %s cipher" % self.encryption)
+            return False
+        if cipher not in ENCRYPTION_CIPHERS:
+            self.warn_and_quit(EXIT_ENCRYPTION, "unsupported server cipher: %s, allowed ciphers: %s" % (cipher, ", ".join(ENCRYPTION_CIPHERS)))
+            return False
+        self._protocol.set_cipher_out(cipher, cipher_iv, self.get_password(), key_salt, iterations)
+
+
+    def get_password(self):
+        if self.password is None:
+            self.load_password()
+        return self.password
 
     def load_password(self):
         try:
@@ -210,8 +272,11 @@ class XpraClientBase(gobject.GObject):
                 self.password = self.password[:-1]
         except IOError, e:
             self.warn_and_quit(EXIT_PASSWORD_FILE_ERROR, "failed to open password file %s: %s" % (self.password_file, e))
+        log("password read from file %s is %s", self.password_file, self.password)
 
     def _process_hello(self, packet):
+        if not self.password_sent and self.password_file:
+            log.warn("Warning: the server did not request our password!")
         self.server_capabilities = packet[1]
         self.parse_server_capabilities(self.server_capabilities)
 
@@ -221,8 +286,11 @@ class XpraClientBase(gobject.GObject):
         if not is_compatible_with(self._remote_version):
             self.warn_and_quit(EXIT_INCOMPATIBLE_VERSION, "incompatible remote version: %s" % self._remote_version)
             return False
-        if capabilities.get("rencode") and has_rencode:
+        if capabilities.get("rencode") and use_rencode:
             self._protocol.enable_rencode()
+        if self.encryption:
+            #server uses a new cipher after second hello:
+            self.set_server_encryption(capabilities)
         self._protocol.chunked_compression = capabilities.get("chunked_compression", False)
         return True
 
@@ -258,8 +326,12 @@ class GLibXpraClient(XpraClientBase):
     def __init__(self, conn, opts):
         XpraClientBase.__init__(self, opts)
         self.ready(conn)
+        gobject.timeout_add(DEFAULT_TIMEOUT, self.timeout)
         self.send_hello()
 
+    def timeout(self, *args):
+        log.warn("timeout!")
+
     def init_packet_handlers(self):
         XpraClientBase.init_packet_handlers(self)
         def noop(*args):
@@ -274,7 +346,7 @@ class GLibXpraClient(XpraClientBase):
             self._packet_handlers[t] = noop
 
     def run(self):
-        import glib
+        glib = import_glib()
         try:
             glib.threads_init()
         except AttributeError:
@@ -310,11 +382,11 @@ class ScreenshotXpraClient(GLibXpraClient):
 
     def __init__(self, conn, opts, screenshot_filename):
         self.screenshot_filename = screenshot_filename
-        def screenshot_timeout(*args):
-            self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the screenshot")
-        gobject.timeout_add(10*1000, screenshot_timeout)
         GLibXpraClient.__init__(self, conn, opts)
 
+    def timeout(self, *args):
+        self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the screenshot")
+
     def _process_screenshot(self, packet):
         (w, h, encoding, _, img_data) = packet[1:6]
         assert encoding=="png"
@@ -341,11 +413,8 @@ class InfoXpraClient(GLibXpraClient):
         it queries the server with an 'info' request
     """
 
-    def __init__(self, conn, opts):
-        def info_timeout(*args):
-            self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the info")
-        gobject.timeout_add(10*1000, info_timeout)
-        GLibXpraClient.__init__(self, conn, opts)
+    def timeout(self, *args):
+        self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the info")
 
     def _process_hello(self, packet):
         log.debug("process_hello: %s", packet)
@@ -368,12 +437,8 @@ class VersionXpraClient(GLibXpraClient):
         it queries the server for version information and prints it out
     """
 
-    def __init__(self, conn, opts):
-        def version_timeout(*args):
-            log.error("timeout: did not receive the version")
-            self.quit(5)
-        gobject.timeout_add(10*1000, version_timeout)
-        GLibXpraClient.__init__(self, conn, opts)
+    def timeout(self, *args):
+        self.warn_and_quit(EXIT_TIMEOUT, "timeout: did not receive the version")
 
     def _process_hello(self, packet):
         log.debug("process_hello: %s", packet)
@@ -390,11 +455,8 @@ class VersionXpraClient(GLibXpraClient):
 class StopXpraClient(GLibXpraClient):
     """ stop a server """
 
-    def __init__(self, conn, opts):
-        def stop_timeout(*args):
-            self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
-        gobject.timeout_add(5*1000, stop_timeout)
-        GLibXpraClient.__init__(self, conn, opts)
+    def timeout(self, *args):
+        self.warn_and_quit(EXIT_TIMEOUT, "timeout: server did not disconnect us")
 
     def _process_hello(self, packet):
-        self.send(["shutdown-server"])
+        gobject.idle_add(self.send, "shutdown-server")
