@@ -97,7 +97,10 @@ else:
 
     WINDOW_POPUP = gtk.WINDOW_POPUP
     WINDOW_TOPLEVEL = gtk.WINDOW_TOPLEVEL
-    WINDOW_EVENT_MASK = gdk.STRUCTURE_MASK | gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK | gdk.POINTER_MOTION_MASK | gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK
+    WINDOW_EVENT_MASK = gdk.STRUCTURE_MASK | gdk.KEY_PRESS_MASK | gdk.KEY_RELEASE_MASK \
+            | gdk.POINTER_MOTION_MASK | gdk.BUTTON_PRESS_MASK | gdk.BUTTON_RELEASE_MASK \
+            | gdk.PROPERTY_CHANGE_MASK
+
     OR_TYPE_HINTS = [gdk.WINDOW_TYPE_HINT_DIALOG,
                 gdk.WINDOW_TYPE_HINT_MENU, gdk.WINDOW_TYPE_HINT_TOOLBAR,
                 #gdk.WINDOW_TYPE_HINT_SPLASHSCREEN, gdk.WINDOW_TYPE_HINT_UTILITY,
@@ -217,6 +220,13 @@ class ClientWindow(gtk.Window):
 
         self.update_metadata(metadata)
 
+        display = gtk.gdk.display_get_default()
+        screen_num = client_properties.get("screen")
+        if screen_num is not None and screen_num>=0 and screen_num<display.get_n_screens():
+            screen = display.get_screen(screen_num)
+            if screen:
+                self.set_screen(screen)
+
         self.set_app_paintable(True)
         self.add_events(WINDOW_EVENT_MASK)
         self.move(x, y)
@@ -227,15 +237,23 @@ class ClientWindow(gtk.Window):
             if transient_for is not None and transient_for.window is not None and type_hint in OR_TYPE_HINTS:
                 transient_for._override_redirect_windows.append(self)
         self.connect("notify::has-toplevel-focus", self._focus_change)
-        #deal with workspace mapping:
+        if CAN_SET_WORKSPACE:
+            self.connect("property-notify-event", self.property_changed)
+
+    def property_changed(self, widget, event):
+        log("property_changed: %s", event.atom)
+        if event.atom=="_NET_WM_DESKTOP" and self._been_mapped and not self._override_redirect:
+            #fake a configure event to send the new client_properties with
+            #the updated workspace number:
+            self.process_configure_event()
 
     def set_workspace(self):
         if not CAN_SET_WORKSPACE or self._been_mapped:
-            return
-        workspace = self._client_properties.get("workspace")
+            return -1
+        workspace = self._client_properties.get("workspace", -1)
         log("set_workspace() workspace=%s", workspace)
-        if not workspace or workspace==self.get_workspace():
-            return
+        if workspace<0 or workspace==self.get_current_workspace():
+            return -1
         try:
             from wimpiggy.lowlevel import sendClientMessage, const  #@UnresolvedImport
             from wimpiggy.error import trap
@@ -243,14 +261,16 @@ class ClientWindow(gtk.Window):
             ndesktops = xget_u32_property(root, "_NET_NUMBER_OF_DESKTOPS")
             log("set_workspace() ndesktops=%s", ndesktops)
             if ndesktops is None or ndesktops<=1:
-                return
+                return  -1
             workspace = max(0, min(ndesktops-1, workspace))
             event_mask = const["SubstructureNotifyMask"] | const["SubstructureRedirectMask"]
             trap.call_synced(sendClientMessage, root, gdk_window(self), False, event_mask, "_NET_WM_DESKTOP",
                       workspace, const["CurrentTime"],
                       0, 0, 0)
+            return workspace
         except Exception, e:
             log.error("failed to set workspace: %s", e)
+            return -1
 
     def is_OR(self):
         return self._override_redirect
@@ -261,15 +281,22 @@ class ClientWindow(gtk.Window):
     def is_GL(self):
         return False
 
-    def get_workspace(self):
-        if sys.platform.startswith("win"):
-            return  -1              #windows does not have workspaces
+    def get_current_workspace(self):
         window = gdk_window(self)
         root = window.get_screen().get_root_window()
-        for target, prop in ((window, "_NET_WM_DESKTOP"), (root, "_NET_CURRENT_DESKTOP")):
-            value = xget_u32_property(target, prop)
-            if value is not None:
-                return value
+        return self.do_get_workspace(root, "_NET_CURRENT_DESKTOP")
+
+    def get_window_workspace(self):
+        return self.do_get_workspace(gdk_window(self), "_NET_WM_DESKTOP")
+
+    def do_get_workspace(self, target, prop):
+        if sys.platform.startswith("win"):
+            return  -1              #windows does not have workspaces
+        value = xget_u32_property(target, prop)
+        if value is not None:
+            log("do_get_workspace() found value=%s from %s / %s", value, target, prop)
+            return value
+        log("do_get_workspace() value not found!")
         return  -1
 
     def new_backing(self, w, h):
@@ -439,14 +466,22 @@ class ClientWindow(gtk.Window):
     def do_map_event(self, event):
         log("Got map event: %s", event)
         gtk.Window.do_map_event(self, event)
-        #set group leader (but avoid ugly "not implemented" warning on win32):
-        if self.group_leader and not sys.platform.startswith("win"):
+        if self.group_leader:
             self.window.set_group(self.group_leader)
-        self.set_workspace()
         if not self._override_redirect:
             x, y, w, h = get_window_geometry(self)
-            client_properties = {"workspace" : self.get_workspace()}
-            self._client.send("map-window", self._id, x, y, w, h, client_properties)
+            if not self._been_mapped:
+                workspace = self.set_workspace()
+            else:
+                #window has been mapped, so these attributes can be read (if present):
+                self._client_properties["screen"] = self.get_screen().get_number()
+                workspace = self.get_window_workspace()
+                if workspace<0:
+                    workspace = self.get_current_workspace()
+            if workspace>=0:
+                self._client_properties["workspace"] = workspace
+            log("map-window for wid=%s with client props=%s", self._id, self._client_properties)
+            self._client.send("map-window", self._id, x, y, w, h, self._client_properties)
             self._pos = (x, y)
             self._size = (w, h)
         self._been_mapped = True
@@ -455,8 +490,10 @@ class ClientWindow(gtk.Window):
     def do_configure_event(self, event):
         log("Got configure event: %s", event)
         gtk.Window.do_configure_event(self, event)
-        if self._override_redirect:
-            return
+        if not self._override_redirect:
+            self.process_configure_event()
+
+    def process_configure_event(self):
         x, y, w, h = get_window_geometry(self)
         w = max(1, w)
         h = max(1, h)
@@ -465,8 +502,16 @@ class ClientWindow(gtk.Window):
         self._pos = (x, y)
         if self._client.window_configure:
             #if we support configure-window, send that first
-            client_properties = {"workspace" : self.get_workspace()}
-            self._client.send("configure-window", self._id, x, y, w, h, client_properties)
+            if self._been_mapped:
+                #if the window has been mapped already, the workspace should be set:
+                self._client_properties["screen"] = self.get_screen().get_number()
+                workspace = self.get_window_workspace()
+                if workspace<0:
+                    workspace = self.get_current_workspace()
+                if workspace>=0:
+                    self._client_properties["workspace"] = workspace
+            log("configure-window for wid=%s with client props=%s", self._id, self._client_properties)
+            self._client.send("configure-window", self._id, x, y, w, h, self._client_properties)
         if dx!=0 or dy!=0:
             #window has moved
             if not self._client.window_configure:
