@@ -1,6 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2012 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2013 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -15,6 +15,11 @@ from subprocess import Popen, PIPE
 import signal
 import shlex
 
+#this is here so we can expose the "platform" module
+#before we import xpra.platform
+import platform as python_platform
+assert python_platform
+
 import xpra
 from xpra.dotxpra import DotXpra
 from xpra.platform import (XPRA_LOCAL_SERVERS_SUPPORTED,
@@ -24,7 +29,7 @@ from xpra.platform import (XPRA_LOCAL_SERVERS_SUPPORTED,
 from xpra.bytestreams import TwoFileConnection, SocketConnection
 
 from wimpiggy.gobject_compat import import_gobject, is_gtk3
-import_gobject()
+gobject = import_gobject()
 try:
     import Image
     assert Image
@@ -82,22 +87,58 @@ if "rgb24" in ENCODINGS:
         pass
     except Exception, e:
         print("cannot load webp: %s" % e)
+ENCRYPTION_CIPHERS = []
+try:
+    from Crypto.Cipher import AES
+    assert AES
+    ENCRYPTION_CIPHERS.append("AES")
+except:
+    pass
 
 
-def read_xpra_conf(conf_dir):
+def get_build_info():
+    info = []
+    try:
+        from xpra.build_info import BUILT_BY, BUILT_ON, BUILD_DATE, REVISION, LOCAL_MODIFICATIONS
+        info.append("Built on %s by %s" % (BUILT_ON, BUILT_BY))
+        if BUILD_DATE:
+            info.append(BUILD_DATE)
+        if int(LOCAL_MODIFICATIONS)==0:
+            info.append("revision %s" % REVISION)
+        else:
+            info.append("revision %s with %s local changes" % (REVISION, LOCAL_MODIFICATIONS))
+    except Exception, e:
+        print("Error: could not find the build information: %s", e)
+    return info
+
+
+def read_config(conf_file):
     d = {}
-    if not os.path.exists(conf_dir) or not os.path.isdir(conf_dir):
-        return  d
-    conf_file = os.path.join(conf_dir, 'xpra.conf')
-    if not os.path.exists(conf_file) or not os.path.isfile(conf_file):
-        return  d
     f = open(conf_file, "rU")
+    lines = []
     for line in f:
-        sline = line.strip()
+        sline = line.strip().rstrip('\r\n').strip()
         if len(sline) == 0:
             continue
         if sline[0] in ( '!', '#' ):
             continue
+        lines.append(sline)
+    f.close()
+    #aggregate any lines with trailing bacakslash
+    agg_lines = []
+    l = ""
+    for line in lines:
+        if line.endswith("\\"):
+            l += line[:-1]
+        else:
+            l += line
+            agg_lines.append(l)
+            l = ""
+    if len(l)>0:
+        #last line had a trailing backslash... meh
+        agg_lines.append(l)
+    #parse name=value pairs:
+    for sline in agg_lines:
         if sline.find("=")<=0:
             continue
         props = sline.split("=", 1)
@@ -112,9 +153,17 @@ def read_xpra_conf(conf_dir):
                 d[name] = [current_value, value]
         else:
             d[name] = value
-    f.close()
     return  d
 
+def read_xpra_conf(conf_dir):
+    d = {}
+    if not os.path.exists(conf_dir) or not os.path.isdir(conf_dir):
+        return  d
+    conf_file = os.path.join(conf_dir, 'xpra.conf')
+    if not os.path.exists(conf_file) or not os.path.isfile(conf_file):
+        return  d
+    return read_config(conf_file)
+
 def read_xpra_defaults():
     #first, read the global defaults:
     if sys.platform.startswith("win"):
@@ -134,6 +183,7 @@ def read_xpra_defaults():
         defaults[k] = v
     return defaults
 
+
 def nox():
     if "DISPLAY" in os.environ:
         del os.environ["DISPLAY"]
@@ -143,6 +193,8 @@ def nox():
     warnings.filterwarnings("error", "could not open display")
 
 def main(script_file, cmdline):
+    if os.name=="posix" and os.getuid()==0:
+        print("\nWarning: running as root")
     #################################################################
     ## NOTE NOTE NOTE
     ##
@@ -175,6 +227,12 @@ def main(script_file, cmdline):
         return parse_or_use_default(varname, default_value, bool_parse)
     def int_default(varname, default_value):
         return parse_or_use_default(varname, default_value, int)
+    def int_auto_default(varname, default_value):
+        def int_or_auto(val):
+            if val=="auto":
+                return default_value
+            return int(val)
+        return parse_or_use_default(varname, default_value, int_or_auto)
     def float_default(varname, default_value):
         return parse_or_use_default(varname, default_value, float)
     def string_list(varname, default_value):
@@ -185,18 +243,23 @@ def main(script_file, cmdline):
             return v
         return [str(v)]
 
-    if XPRA_LOCAL_SERVERS_SUPPORTED:
+    supports_server = XPRA_LOCAL_SERVERS_SUPPORTED
+    if supports_server:
+        try:
+            from xpra.wait_for_x_server import wait_for_x_server    #@UnresolvedImport @UnusedImport
+        except:
+            supports_server = False
+
+    if supports_server:
         start_str = "\t%prog start DISPLAY\n"
         list_str = "\t%prog list\n"
-        upgrade_str = "\t%prog upgrade DISPLAY"
+        upgrade_str = "\t%prog upgrade DISPLAY\n"
+        shadow_str = "\t%prog shadow DISPLAY\n"
         note_str = ""
         stop_str = "\t%prog stop [DISPLAY]\n"
     else:
-        start_str = ""
-        list_str = ""
-        upgrade_str = ""
+        start_str, list_str, upgrade_str, shadow_str, stop_str = "", "", "", "", ""
         note_str = "(This xpra installation does not support starting local servers.)"
-        stop_str = ""
     parser = OptionParser(version="xpra v%s" % xpra.__version__,
                           usage="".join(["\n",
                                          start_str,
@@ -208,8 +271,9 @@ def main(script_file, cmdline):
                                          stop_str,
                                          list_str,
                                          upgrade_str,
+                                         shadow_str,
                                          note_str]))
-    if XPRA_LOCAL_SERVERS_SUPPORTED:
+    if supports_server:
         group = OptionGroup(parser, "Server Options",
                     "These options are only relevant on the server when using the 'start' or 'upgrade' mode.")
         group.add_option("--start-child", action="append",
@@ -234,21 +298,27 @@ def main(script_file, cmdline):
                           dest="bind_tcp", default=None,
                           metavar="[HOST]:PORT",
                           help="Listen for connections over TCP (insecure)")
+        group.add_option("--no-pulseaudio", action="store_false",
+                      dest="pulseaudio", default=bool_default("pulseaudio", True),
+                      help="Disable starting of a pulseaudio server for the session")
+        group.add_option("--pulseaudio-command", action="store",
+                      dest="pulseaudio_command", default=defaults.get("pulseaudio-command", ""),
+                      help="The command used to start the pulseaudio server (default: '%default')")
         parser.add_option_group(group)
 
     group = OptionGroup(parser, "Server Controlled Features",
-                "These options can be used to turn on or off certain features, "
+                "These options can be used to turn certain features on or off, "
                 "they can be specified on the client or on the server, "
                 "but the client cannot enable them if they are disabled on the server.")
     group.add_option("--no-clipboard", action="store_false",
                       dest="clipboard", default=bool_default("clipboard", True),
                       help="Disable clipboard support")
-    group.add_option("--no-pulseaudio", action="store_false",
-                      dest="pulseaudio", default=bool_default("pulseaudio", True),
-                      help="Disable pulseaudio support via X11 root window properties")
     group.add_option("--no-notifications", action="store_false",
                       dest="notifications", default=bool_default("notifications", True),
                       help="Disable forwarding of system notifications")
+    group.add_option("--no-system-tray", action="store_false",
+                      dest="system_tray", default=bool_default("system-tray", True),
+                      help="Disable forwarding of system tray icons")
     group.add_option("--no-cursors", action="store_false",
                       dest="cursors", default=bool_default("cursors", True),
                       help="Disable forwarding of custom application mouse cursors")
@@ -264,30 +334,56 @@ def main(script_file, cmdline):
     group.add_option("--enable-sharing", action="store_true",
                       dest="sharing", default=bool_default("sharing", False),
                       help="Allow more than one client to connect to the same session")
+    group.add_option("--no-speaker", action="store_false",
+                      dest="speaker", default=bool_default("speaker", True),
+                      help="Disable forwarding of sound output to the client(s)")
+    group.add_option("--speaker-codec", action="append",
+                      dest="speaker_codec", default=string_list("speaker-codec", []),
+                      help="The audio codec to use for forwarding the speaker sound output "
+                      "(you may specify more than one to define the preferred order, use 'help' to get a list of options, "
+                      "when unspecified all available codecs are allowed and the first one is used)")
+    group.add_option("--no-microphone", action="store_false",
+                      dest="microphone", default=bool_default("microphone", True),
+                      help="Disable forwarding of sound input to the server")
+    group.add_option("--microphone-codec", action="append",
+                      dest="microphone_codec", default=string_list("microphone-codec", []),
+                      help="The audio codec to use for forwaring the microphone sound input "
+                      "(you may specify more than one to define the preferred order, use 'help' to get a list of options, "
+                      "when unspecified all available codecs are allowed and the first one is used)")
     parser.add_option_group(group)
 
     group = OptionGroup(parser, "Client Picture Encoding and Compression Options",
-                "These options are used by the client to specify the desired picture and network data compression.")
+                "These options are used by the client to specify the desired picture and network data compression."
+                "They may also be specified on the server as default values for those clients that do not set them.")
     default_encoding = defaults.get("encoding", ENCODINGS[0])
     group.add_option("--encoding", action="store",
                       metavar="ENCODING", default=default_encoding,
                       dest="encoding", type="str",
                       help="What image compression algorithm to use: %s. Default: %s" % (", ".join(ENCODINGS), default_encoding))
     if "jpeg" in ENCODINGS:
-        group.add_option("--jpeg-quality", action="store",
-                          metavar="LEVEL",
-                          dest="quality", type="int", default=None,
-                          help="Deprecated - use 'quality'")
         group.add_option("-b", "--max-bandwidth", action="store",
                           dest="max_bandwidth", type="float", default=0.0, metavar="BANDWIDTH (kB/s)",
                           help="Specify the link's maximal receive speed to auto-adjust JPEG quality, 0.0 disables. (default: disabled)")
     if len(set(("jpeg", "webp", "x264")).intersection(set(ENCODINGS)))>0:
+        group.add_option("--min-quality", action="store",
+                          metavar="MIN-LEVEL",
+                          dest="min_quality", type="int", default=int_default("min-quality", 0),
+                          help="Sets the minimum x264 encoding quality allowed in automatic quality setting (from 1 to 100, 0 to leave unset). Default: %default.")
         group.add_option("--quality", action="store",
                           metavar="LEVEL",
-                          dest="quality", type="int", default=int_default("quality", -1),
-                          help="Use image compression with the given quality - only relevant to lossy encodings (1-100, -1 to use automatic settings). Default: %default.")
+                          dest="quality", type="int", default=int_auto_default("quality", 0),
+                          help="Use a fixed image compression quality - only relevant to lossy encodings (1-100, 0 to use automatic setting). Default: %default.")
+    if "x264" in ENCODINGS:
+        group.add_option("--min-speed", action="store",
+                          metavar="SPEED",
+                          dest="min_speed", type="int", default=int_default("min-speed", 0),
+                          help="Sets the minimum x264 encoding speed allowed in automatic speed setting (1-100, 0 to leave unset). Default: %default.")
+        group.add_option("--speed", action="store",
+                          metavar="SPEED",
+                          dest="speed", type="int", default=int_auto_default("speed", 0),
+                          help="Use x264 image compression with the given encoding speed (1-100, 0 to use automatic setting). Default: %default.")
     group.add_option("--auto-refresh-delay", action="store",
-                      dest="auto_refresh_delay", type="float", default=float_default("auto-refresh-delay", 1.0),
+                      dest="auto_refresh_delay", type="float", default=float_default("auto-refresh-delay", 0.25),
                       metavar="DELAY",
                       help="Idle delay in seconds before doing an automatic lossless refresh."
                       + " 0.0 to disable."
@@ -374,23 +470,42 @@ def main(script_file, cmdline):
                       dest="remote_xpra", default=".xpra/run-xpra",
                       metavar="CMD",
                       help="How to run xpra on the remote host (default: '%default')")
+    if len(ENCRYPTION_CIPHERS)>0:
+        group.add_option("--encryption", action="store",
+                          dest="encryption", default=None,
+                          metavar="CMD",
+                          help="Specifies the encryption cipher to use, only %s is currently supported. (default: None)" % (", ".join(ENCRYPTION_CIPHERS)))
     parser.add_option_group(group)
 
-    (options, args) = parser.parse_args(cmdline[1:])
+    options, args = parser.parse_args(cmdline[1:])
+    if not args:
+        parser.error("need a mode")
+    #ensure the default values are set even though
+    #the option is not shown to the user as it is not available
     if "jpeg" not in ENCODINGS:
-        #ensure the default values are set even though
-        #the option is not shown to the user as it is not available
-        options.quality = 80
         options.max_bandwidth = 0
+    if len(set(("jpeg", "webp", "x264")).intersection(set(ENCODINGS)))==0:
+        options.min_quality = int_default("min-quality", 0)
+        options.quality = int_auto_default("quality", 0)
+    if "x264" not in ENCODINGS:
+        options.min_speed = int_default("min-speed", 0)
+        options.speed = int_auto_default("speed", 0)
     try:
         int(options.dpi)
     except Exception, e:
         parser.error("invalid dpi: %s" % e)
-
-    if not args:
-        parser.error("need a mode")
+    if len(ENCRYPTION_CIPHERS)==0:
+        #if no encryption options are available the option is not shown
+        #but we still need to set a value to avoid errors later on:
+        options.encryption = ""
     if options.encoding and options.encoding not in ENCODINGS:
         parser.error("encoding %s is not supported, try: %s" % (options.encoding, ", ".join(ENCODINGS)))
+    if options.encryption:
+        assert len(ENCRYPTION_CIPHERS)>0
+        if options.encryption not in ENCRYPTION_CIPHERS:
+            parser.error("encryption %s is not supported, try: %s" % (options.encryption, ", ".join(ENCRYPTION_CIPHERS)))
+        if not options.password_file:
+            parser.error("encryption %s cannot be used without a password (see --password-file option)" % options.encryption)
 
     def toggle_logging(level):
         if not options.debug:
@@ -418,7 +533,10 @@ def main(script_file, cmdline):
 
     #configure default logging handler:
     mode = args.pop(0)
-    if mode in ("start", "upgrade", "attach"):
+    if mode in ("start", "upgrade", "attach", "shadow"):
+        if show_codec_help("attach" not in cmdline[1:],
+                           options.speaker_codec, options.microphone_codec):
+            return 0
         logging.basicConfig(format="%(asctime)s %(message)s")
     else:
         logging.root.addHandler(logging.StreamHandler(sys.stdout))
@@ -439,28 +557,78 @@ def main(script_file, cmdline):
         signal.signal(signal.SIGUSR1, sigusr1)
         signal.signal(signal.SIGUSR2, sigusr2)
 
-    if mode in ("start", "upgrade") and XPRA_LOCAL_SERVERS_SUPPORTED:
+    if mode in ("start", "upgrade", "shadow") and supports_server:
         nox()
         from xpra.scripts.server import run_server
         return run_server(parser, options, mode, script_file, args)
     elif mode in ("attach", "detach", "screenshot", "version", "info"):
         return run_client(parser, options, args, mode)
-    elif mode == "stop" and XPRA_LOCAL_SERVERS_SUPPORTED:
+    elif mode == "stop" and supports_server:
         nox()
         return run_stop(parser, options, args)
-    elif mode == "list" and XPRA_LOCAL_SERVERS_SUPPORTED:
+    elif mode == "list" and supports_server:
         return run_list(parser, options, args)
-    elif mode == "_proxy" and XPRA_LOCAL_SERVERS_SUPPORTED:
+    elif mode == "_proxy" and supports_server:
         nox()
         return run_proxy(parser, options, args)
     else:
         parser.error("invalid mode '%s'" % mode)
         return 1
 
+# we end up initializing gstreamer here and it does things
+# we don't want with sys.argv, so hack around it:
+saved_args = sys.argv
+sys.argv = sys.argv[:1]
+try:
+    from xpra.sound import gstreamer_util   #@UnusedImport
+    HAS_SOUND = True
+except:
+    HAS_SOUND = False
+sys.argv = saved_args
+
+def get_codecs(is_speaker, is_server):
+    if not HAS_SOUND:
+        return []
+    try:
+        from xpra.sound.gstreamer_util import can_encode, can_decode
+        if (is_server and is_speaker) or (not is_server and not is_speaker):
+            return can_encode()
+        else:
+            return can_decode()
+    except Exception, e:
+        print("failed to get list of codecs: %s" % e)
+        return []
+
+def show_codec_help(is_server, speaker_codecs, microphone_codecs):
+    all_speaker_codecs = get_codecs(True, is_server)
+    invalid_sc = [x for x in speaker_codecs if x not in all_speaker_codecs]
+    hs = "help" in speaker_codecs
+    if hs:
+        print("speaker codecs available: %s" % (", ".join(all_speaker_codecs)))
+    elif len(invalid_sc):
+        print("WARNING: some of the specified speaker codecs are not available: %s" % (", ".join(invalid_sc)))
+        for x in invalid_sc:
+            speaker_codecs.remove(x)
+    elif len(speaker_codecs)==0:
+        speaker_codecs += all_speaker_codecs
+
+    all_microphone_codecs = get_codecs(True, is_server)
+    invalid_mc = [x for x in microphone_codecs if x not in all_microphone_codecs]
+    hm = "help" in microphone_codecs
+    if hm:
+        print("microphone codecs available: %s" % (", ".join(all_microphone_codecs)))
+    elif len(invalid_mc):
+        print("WARNING: some of the specified microphone codecs are not available: %s" % (", ".join(invalid_mc)))
+        for x in invalid_mc:
+            microphone_codecs.remove(x)
+    elif len(microphone_codecs)==0:
+        microphone_codecs += all_microphone_codecs
+    return hm or hs
+
 def get_default_socket_dir():
     return os.environ.get("XPRA_SOCKET_DIR", "~/.xpra")
 
-def parse_display_name(parser, opts, display_name):
+def parse_display_name(error_cb, opts, display_name):
     desc = {"display_name" : display_name}
     if display_name.startswith("ssh:") or display_name.startswith("ssh/"):
         separator = display_name[3] # ":" or "/"
@@ -468,22 +636,43 @@ def parse_display_name(parser, opts, display_name):
         desc["local"] = False
         parts = display_name.split(separator)
         if len(parts)>2:
-            desc["host"] = separator.join(parts[1:-1])
-            desc["display"] = parts[-1]
-            desc["display"] = ":" + desc["display"]
+            host = separator.join(parts[1:-1])
+            desc["display"] = ":" + parts[-1]
             desc["display_as_args"] = [desc["display"]]
         else:
-            desc["host"] = parts[1]
+            host = parts[1]
             desc["display"] = None
             desc["display_as_args"] = []
         desc["ssh"] = shlex.split(opts.ssh)
-        desc["full_ssh"] = desc["ssh"] + ["-T", desc["host"]]
+        full_ssh = desc["ssh"]
+        upos = host.find("@")
+        if upos>=0:
+            username = host[:upos]
+            host = host[upos+1:]
+            ppos = username.find(":")
+            if ppos>=0:
+                password = username[ppos+1:]
+                username = username[:ppos]
+                desc["password"] = password
+            desc["username"] = username
+            full_ssh += ["-l", username]
+        desc["host"] = host
+        full_ssh += ["-T", desc["host"]]
+        desc["full_ssh"] = full_ssh
         remote_xpra = opts.remote_xpra.split()
         if opts.sockdir:
             #ie: XPRA_SOCKET_DIR=/tmp .xpra/run-xpra _proxy :10
             remote_xpra.append("--socket-dir=%s" % opts.sockdir)
         desc["remote_xpra"] = remote_xpra
-        desc["full_remote_xpra"] = desc["full_ssh"] + desc["remote_xpra"]
+        if desc.get("password") is None and opts.password_file and os.path.exists(opts.password_file):
+            try:
+                try:
+                    passwordFile = open(opts.password_file, "rb")
+                    desc["password"] = passwordFile.read()
+                finally:
+                    passwordFile.close()
+            except Exception, e:
+                print("failed to read password file %s: %s", opts.password_file, e)
         return desc
     elif display_name.startswith(":"):
         desc["type"] = "unix-domain"
@@ -498,14 +687,14 @@ def parse_display_name(parser, opts, display_name):
         parts = display_name.split(separator)
         port = int(parts[-1])
         if port<=0 or port>=65536:
-            parser.error("invalid port number: %s" % port)
+            error_cb("invalid port number: %s" % port)
         desc["port"] = port
         desc["host"] = separator.join(parts[1:-1])
         if desc["host"] == "":
             desc["host"] = "127.0.0.1"
         return desc
     else:
-        parser.error("unknown format for display name: %s" % display_name)
+        error_cb("unknown format for display name: %s" % display_name)
 
 def pick_display(parser, opts, extra_args):
     if len(extra_args) == 0:
@@ -518,37 +707,54 @@ def pick_display(parser, opts, extra_args):
         if len(live_servers) == 0:
             parser.error("cannot find a live server to connect to")
         elif len(live_servers) == 1:
-            return parse_display_name(parser, opts, live_servers[0])
+            return parse_display_name(parser.error, opts, live_servers[0])
         else:
             parser.error("there are multiple servers running, please specify")
     elif len(extra_args) == 1:
-        return parse_display_name(parser, opts, extra_args[0])
+        return parse_display_name(parser.error, opts, extra_args[0])
     else:
         parser.error("too many arguments")
 
 def _socket_connect(sock, endpoint, description):
-    try:
-        sock.settimeout(10)
-        sock.connect(endpoint)
-    except socket.error, e:
-        sys.exit("Connection failed: %s" % (e,))
-        return
+    sock.connect(endpoint)
     sock.settimeout(None)
     return SocketConnection(sock, sock.getsockname(), sock.getpeername(), description)
 
 def connect_or_fail(display_desc):
+    try:
+        return connect_to(display_desc)
+    except Exception, e:
+        sys.exit("connection failed: %s" % str(e))
+
+def connect_to(display_desc, debug_cb=None):
     display_name = display_desc["display_name"]
     if display_desc["type"] == "ssh":
-        cmd = (display_desc["full_remote_xpra"]
-               + ["_proxy"] + display_desc["display_as_args"])
+        cmd = display_desc["full_ssh"]
+        if sys.platform.startswith("win"):
+            password = display_desc.get("password")
+            if password:
+                cmd += ["-pw", password]
+            cmd.append("-ssh")
+            cmd.append("-agent")
+        cmd += display_desc["remote_xpra"] + ["_proxy"] + display_desc["display_as_args"]
         try:
-            def setsid():
-                #run in a new session
-                if os.name=="posix":
+            kwargs = {}
+            if os.name=="posix" and not sys.platform.startswith("darwin"):
+                def setsid():
+                    #run in a new session
                     os.setsid()
-            child = Popen(cmd, stdin=PIPE, stdout=PIPE, preexec_fn=setsid)
+                kwargs["preexec_fn"] = setsid
+            elif sys.platform.startswith("win"):
+                from subprocess import CREATE_NEW_PROCESS_GROUP, CREATE_NEW_CONSOLE
+                flags = CREATE_NEW_PROCESS_GROUP | CREATE_NEW_CONSOLE
+                kwargs["creationflags"] = flags
+                kwargs["stderr"] = PIPE
+            if debug_cb:
+                debug_cb("starting %s tunnel" % str(cmd[0]))
+                #debug_cb("starting ssh: %s with kwargs=%s" % (str(cmd), kwargs))
+            child = Popen(cmd, stdin=PIPE, stdout=PIPE, **kwargs)
         except OSError, e:
-            sys.exit("Error running ssh program '%s': %s" % (cmd[0], e))
+            raise Exception("Error running ssh program '%s': %s" % (cmd, e))
         def abort_test(action):
             """ if ssh dies, we don't need to try to read/write from its sockets """
             e = child.poll()
@@ -558,22 +764,48 @@ def connect_or_fail(display_desc):
                 from wimpiggy.util import gtk_main_quit_really
                 gtk_main_quit_really()
                 raise IOError(error_message)
-        return TwoFileConnection(child.stdin, child.stdout, abort_test, target=display_name)
+        def stop_tunnel():
+            try:
+                if child.poll() is None:
+                    #only supported on win32 since Python 2.7
+                    if hasattr(child, "terminate"):
+                        child.terminate()
+                    elif hasattr(os, "kill"):
+                        os.kill(child.pid, signal.SIGTERM)
+                    else:
+                        raise Exception("cannot find function to kill subprocess")
+            except Exception, e:
+                print("error trying to stop ssh tunnel process: %s" % e)
+        return TwoFileConnection(child.stdin, child.stdout, abort_test, target=display_name, info="SSH", close_cb=stop_tunnel)
 
     elif display_desc["type"] == "unix-domain":
         sockdir = DotXpra(display_desc["sockdir"])
         sock = socket.socket(socket.AF_UNIX)
+        sock.settimeout(5)
         sockfile = sockdir.socket_path(display_desc["display"])
         return _socket_connect(sock, sockfile, display_name)
 
     elif display_desc["type"] == "tcp":
         sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        sock.settimeout(10)
         tcp_endpoint = (display_desc["host"], display_desc["port"])
         return _socket_connect(sock, tcp_endpoint, display_name)
 
     else:
         assert False, "unsupported display type in connect"
 
+def set_signal_handlers(app):
+    def deadly_signal(signum, frame):
+        print("got deadly signal %s, exiting" % {signal.SIGINT:"SIGINT", signal.SIGTERM:"SIGTERM"}.get(signum, signum))
+        app.cleanup()
+        os._exit(128 + signum)
+    def app_signal(signum, frame):
+        print("\ngot signal %s, exiting" % {signal.SIGINT:"SIGINT", signal.SIGTERM:"SIGTERM"}.get(signum, signum))
+        signal.signal(signal.SIGINT, deadly_signal)
+        signal.signal(signal.SIGTERM, deadly_signal)
+        gobject.timeout_add(0, app.quit, 128 + signum, priority=gobject.PRIORITY_HIGH)
+    signal.signal(signal.SIGINT, app_signal)
+    signal.signal(signal.SIGTERM, app_signal)
 
 def run_client(parser, opts, extra_args, mode):
     if mode=="screenshot":
@@ -582,7 +814,7 @@ def run_client(parser, opts, extra_args, mode):
         screenshot_filename = extra_args[0]
         extra_args = extra_args[1:]
 
-    if mode in ("info", "attach"):
+    if mode in ("attach"):
         sys.stdout.write("xpra client version %s\n" % xpra.__version__)
         sys.stdout.flush()
 
@@ -623,10 +855,7 @@ def run_client(parser, opts, extra_args, mode):
         elif mode=="attach":
             log.info("Attached to %s (press Control-C to detach)\n" % conn.target)
     app.connect("handshake-complete", handshake_complete)
-    def client_SIGINT(*args):
-        print("received SIGINT, closing")
-        app.quit(-signal.SIGINT)
-    signal.signal(signal.SIGINT, client_SIGINT)
+    set_signal_handlers(app)
     try:
         try:
             return app.run()
@@ -639,7 +868,9 @@ def run_proxy(parser, opts, extra_args):
     from xpra.proxy import XpraProxy
     assert "gtk" not in sys.modules
     server_conn = connect_or_fail(pick_display(parser, opts, extra_args))
-    app = XpraProxy(TwoFileConnection(sys.stdout, sys.stdin), server_conn)
+    app = XpraProxy(TwoFileConnection(sys.stdout, sys.stdin, info="stdin/stdout"), server_conn)
+    signal.signal(signal.SIGINT, app.quit)
+    signal.signal(signal.SIGTERM, app.quit)
     app.run()
     return  0
 
@@ -681,6 +912,17 @@ def run_stop(parser, opts, extra_args):
     return  e
 
 
+def may_cleanup_socket(sockdir, state, display, clean_states=[DotXpra.DEAD]):
+    sys.stdout.write("\t%s session at %s" % (state, display))
+    if state in clean_states:
+        try:
+            os.unlink(sockdir.socket_path(display))
+        except OSError:
+            pass
+        else:
+            sys.stdout.write(" (cleaned up)")
+    sys.stdout.write("\n")
+
 def run_list(parser, opts, extra_args):
     assert "gtk" not in sys.modules
     if extra_args:
@@ -692,15 +934,30 @@ def run_list(parser, opts, extra_args):
         return  1
     sys.stdout.write("Found the following xpra sessions:\n")
     for state, display in results:
-        sys.stdout.write("\t%s session at %s" % (state, display))
-        if state is DotXpra.DEAD:
-            try:
-                os.unlink(sockdir.socket_path(display))
-            except OSError:
-                pass
+        may_cleanup_socket(sockdir, state, display)
+    #now, re-probe the "unknown" ones:
+    unknown = [display for state, display in results if state==DotXpra.UNKNOWN]
+    if len(unknown)>0:
+        sys.stdout.write("Re-probing unknown sessions: %s\n" % (", ".join(unknown)))
+    counter = 0
+    while len(unknown)>0 and counter<5:
+        time.sleep(1)
+        counter += 1
+        probe_list = list(unknown)
+        unknown = []
+        for display in probe_list:
+            state = sockdir.server_state(display)
+            if state is DotXpra.DEAD:
+                may_cleanup_socket(sockdir, state, display)
+            elif state is DotXpra.UNKNOWN:
+                unknown.append(display)
             else:
-                sys.stdout.write(" (cleaned up)")
-        sys.stdout.write("\n")
+                sys.stdout.write("\t%s session at %s\n" % (state, display))
+    #now cleanup those still unknown:
+    clean_states = [DotXpra.DEAD, DotXpra.UNKNOWN]
+    for display in unknown:
+        state = sockdir.server_state(display)
+        may_cleanup_socket(sockdir, state, display, clean_states=clean_states)
     return 0
 
 if __name__ == "__main__":
