@@ -1,5 +1,5 @@
 # This file is part of Parti.
-# Copyright (C) 2010, 2012 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2013 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -38,7 +38,7 @@ def run_cleanups():
             traceback.print_exception(*sys.exc_info())
 
 def deadly_signal(signum, frame):
-    print("got signal %s, exiting" % signum)
+    print("got deadly signal %s, exiting" % {signal.SIGINT:"SIGINT", signal.SIGTERM:"SIGTERM"}.get(signum, signum))
     run_cleanups()
     # This works fine in tests, but for some reason if I use it here, then I
     # get bizarre behavior where the signal handler runs, and then I get a
@@ -52,34 +52,33 @@ def deadly_signal(signum, frame):
 # child to exit and us to receive the SIGCHLD before our fork() returns (and
 # thus before we even know the pid of the child).  So be careful:
 class ChildReaper(object):
-    def __init__(self, app, exit_with_children):
-        self._app = app
-        self._children_pids = None
-        self._dead_pids = set()
-        self._exit_with_children = exit_with_children
-
-    def set_children_pids(self, children_pids):
-        assert self._children_pids is None
+    def __init__(self, quit_cb, children_pids):
+        self._quit = quit_cb
         self._children_pids = children_pids
-        self.check()
+        self._dead_pids = set()
 
     def check(self):
         if (self._children_pids
-            and self._exit_with_children
             and self._children_pids.issubset(self._dead_pids)):
-            print("all children have exited and --exit-with-children was specified, exiting")
-            self._app.quit(False)
+            self._quit()
 
-    def __call__(self, signum, frame):
-        while 1:
+    def sigchld(self, signum, frame):
+        self.reap()
+
+    def add_dead_pid(self, pid):
+        if pid not in self._dead_pids:
+            self._dead_pids.add(pid)
+            self.check()
+
+    def reap(self):
+        while True:
             try:
                 pid, _ = os.waitpid(-1, os.WNOHANG)
             except OSError:
                 break
             if pid == 0:
                 break
-            self._dead_pids.add(pid)
-            self.check()
+            self.add_dead_pid(pid)
 
 def save_pid(pid):
     import gtk
@@ -194,12 +193,24 @@ def close_all_fds(exceptions=[]):
 def run_server(parser, opts, mode, xpra_file, extra_args):
     if len(extra_args) != 1:
         parser.error("need exactly 1 extra argument")
+    assert mode in ("start", "upgrade", "shadow")
+    upgrading = mode == "upgrade"
+    shadowing = mode == "shadow"
     display_name = extra_args.pop(0)
-    if display_name==":0" or display_name.startswith(":0."):
-        sys.stderr.write("WARNING:")
-        sys.stderr.write("You are attempting to run the xpra server against the default X11 display '%s'." % display_name)
-        sys.stderr.write("This is generally not what you want.")
-        sys.stderr.write("You should probably use a higher display number just to avoid any confusion (and also this warning message).")
+    if display_name.startswith(":") and not shadowing:
+        n = display_name[1:]
+        p = n.find(".")
+        if p>0:
+            n = n[:p]
+        try:
+            dno = int(n)
+            if dno>=0 and dno<10:
+                sys.stderr.write("WARNING: low display number: %s\n" % dno)
+                sys.stderr.write("You are attempting to run the xpra server against what seems to be a default X11 display '%s'.\n" % display_name)
+                sys.stderr.write("This is generally not what you want.\n")
+                sys.stderr.write("You should probably use a higher display number just to avoid any confusion (and also this warning message).\n")
+        except:
+            pass
 
     if opts.exit_with_children and not opts.children:
         sys.stderr.write("--exit-with-children specified without any children to spawn; exiting immediately")
@@ -209,9 +220,6 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
     signal.signal(signal.SIGINT, deadly_signal)
     signal.signal(signal.SIGTERM, deadly_signal)
 
-    assert mode in ("start", "upgrade")
-    upgrading = (mode == "upgrade")
-
     from xpra.scripts.main import get_default_socket_dir
     dotxpra = DotXpra(opts.sockdir or get_default_socket_dir())
 
@@ -231,7 +239,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
 
     # Daemonize:
     if opts.daemon:
-        logpath = dotxpra.conf_path(display_name) + ".log"
+        logpath = dotxpra.log_path(display_name) + ".log"
         sys.stderr.write("Entering daemon mode; "
                          + "any further errors will be reported to:\n"
                          + ("  %s\n" % logpath))
@@ -284,7 +292,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
             tcp_socket = create_tcp_socket(parser, opts.bind_tcp)
             sockets.append(tcp_socket)
             def cleanup_tcp_socket():
-                log.info("closing tcp socket %s" % opts.bind_tcp)
+                log.info("closing tcp socket %s", opts.bind_tcp)
                 try:
                     tcp_socket.close()
                 except:
@@ -296,7 +304,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
     #print("creating server socket %s" % sockpath)
     clobber = upgrading or opts.use_display
     try:
-        sockpath = dotxpra.server_socket_path(display_name, clobber)
+        sockpath = dotxpra.server_socket_path(display_name, clobber, wait_for_unknown=5)
     except ServerSockInUse:
         parser.error("You already have an xpra server running at %s\n"
                      "  (did you want 'xpra upgrade'?)"
@@ -313,7 +321,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
     # Do this after writing out the shell script:
     os.environ["DISPLAY"] = display_name
 
-    if not clobber:
+    if not clobber and not shadowing:
         # We need to set up a new server environment
         xauthority = os.environ.get("XAUTHORITY", os.path.expanduser("~/.Xauthority"))
         subs = {"XAUTHORITY" : xauthority,
@@ -327,8 +335,12 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
         xvfb_cmd = xvfb.split()
         xvfb_executable = xvfb_cmd[0]
         xvfb_cmd[0] = "%s-for-Xpra-%s" % (xvfb_executable, display_name)
+        def setsid():
+            #run in a new session
+            if os.name=="posix":
+                os.setsid()
         try:
-            xvfb = subprocess.Popen(xvfb_cmd+[display_name], executable=xvfb_executable, close_fds=True)
+            xvfb = subprocess.Popen(xvfb_cmd+[display_name], executable=xvfb_executable, close_fds=True, preexec_fn=setsid)
         except OSError, e:
             sys.stderr.write("Error starting Xvfb: %s\n" % (e,))
             return  1
@@ -343,7 +355,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
             sys.stderr.write("Error running \"%s\": %s\n" % (" ".join(xauth_cmd), e))
 
     def xvfb_error(instance_exists=False):
-        if clobber:
+        if clobber or shadowing:
             return False
         if xvfb.poll() is None:
             return False
@@ -378,7 +390,9 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
         default_display.close()
     manager.set_default_display(display)
 
-    if clobber:
+    if shadowing:
+        xvfb_pid = None
+    elif clobber:
         xvfb_pid = get_pid()
     else:
         if xvfb_error(True):
@@ -393,51 +407,100 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
         for display in gtk.gdk.display_manager_get().list_displays():
             display.close()
         os.kill(xvfb_pid, signal.SIGTERM)
-    if xvfb_pid is not None and not opts.use_display:
+    if xvfb_pid is not None and not opts.use_display and not shadowing:
         _cleanups.append(kill_xvfb)
 
-    try:
-        from wimpiggy.lowlevel import displayHasXComposite     #@UnresolvedImport
-        from xpra.server import XpraServer
-    except ImportError, e:
-        log.error("Failed to load Xpra server components, check your installation: %s" % e)
-        return 1
-    root = gtk.gdk.get_default_root_window()
-    if not displayHasXComposite(root):
-        log.error("Xpra is a compositing manager, it cannot use a display which lacks the XComposite extension!")
-        return 1
+    if shadowing:
+        from xpra.shadow_server import XpraShadowServer
+        app = XpraShadowServer(sockets, opts)
+    else:
+        try:
+            from wimpiggy.lowlevel import displayHasXComposite     #@UnresolvedImport
+            # This import is delayed because the module depends on gtk:
+            from xpra.server import XpraServer
+        except ImportError, e:
+            log.error("Failed to load Xpra server components, check your installation: %s" % e)
+            return 1
+        root = gtk.gdk.get_default_root_window()
+        if not displayHasXComposite(root):
+            log.error("Xpra is a compositing manager, it cannot use a display which lacks the XComposite extension!")
+            return 1
 
-    # This import is delayed because the module depends on gtk:
-    app = XpraServer(clobber, sockets, opts)
+        app = XpraServer(clobber, sockets, opts)
 
-    child_reaper = ChildReaper(app, opts.exit_with_children)
-    # Always register the child reaper, because even if exit_with_children is
-    # false, we still need to reap them somehow to avoid zombies:
-    signal.signal(signal.SIGCHLD, child_reaper)
+    children_pids = set()
+    def reaper_quit():
+        if opts.exit_with_children:
+            log.info("all children have exited and --exit-with-children was specified, exiting")
+            app.quit(False)
+    child_reaper = ChildReaper(reaper_quit, children_pids)
+    procs = []
+    if sys.version_info < (2, 7) or sys.version_info[:2] == (3, 0):
+        POLL_DELAY = int(os.environ.get("XPRA_POLL_DELAY", 2))
+        log.warn("Warning: outdated/buggy version of Python (%s), switching to process polling every %s seconds to support 'exit-with-children'", ".".join(str(x) for x in sys.version_info), POLL_DELAY)
+        ChildReaper.processes = procs
+        def check_procs():
+            for proc in ChildReaper.processes:
+                if proc.poll() is not None:
+                    child_reaper.add_dead_pid(proc.pid)
+                    child_reaper.check()
+            ChildReaper.processes = [proc for proc in ChildReaper.processes if proc.poll() is None]
+            return True
+        gobject.timeout_add(POLL_DELAY*1000, check_procs)
+    else:
+        #with non-buggy python, we can just check the list of pids
+        #whenever we get a SIGCHLD
+        signal.signal(signal.SIGCHLD, child_reaper.sigchld)
+        # Check once after the mainloop is running, just in case the exit
+        # conditions are satisfied before we even enter the main loop.
+        # (Programming with unix the signal API sure is annoying.)
+        def check_once():
+            child_reaper.check()
+            return False # Only call once
+        gobject.timeout_add(0, check_once)
+
+    if not upgrading and not shadowing and opts.pulseaudio and len(opts.pulseaudio_command)>0:
+        pa_proc = subprocess.Popen(opts.pulseaudio_command, shell=True, close_fds=True)
+        procs.append(pa_proc)
+        log.info("pulseaudio server started with pid %s", pa_proc.pid)
+        def check_pa_start():
+            if pa_proc.poll() is not None or pa_proc.pid in child_reaper._dead_pids:
+                log.warn("Warning: pulseaudio has terminated. Either fix the pulseaudio command line or use --no-pulseaudio to avoid this warning.")
+            return False
+        gobject.timeout_add(1000*2, check_pa_start)
+        def cleanup_pa():
+            if pa_proc.poll() is None and pa_proc.pid not in child_reaper._dead_pids:
+                log.info("stopping pulseaudio with pid %s", pa_proc.pid)
+                try:
+                    pa_proc.terminate()
+                except:
+                    pass
+        _cleanups.append(cleanup_pa)
     if opts.exit_with_children:
         assert opts.children
     if opts.children:
+        #disable ubuntu's global menu using env vars:
         env = os.environ.copy()
         env["UBUNTU_MENUPROXY"] = ""
-        children_pids = set()
+        env["QT_X11_NO_NATIVE_MENUBAR"] = "1"
         for child_cmd in opts.children:
             if child_cmd:
                 try:
-                    children_pids.add(subprocess.Popen(child_cmd, env=env, shell=True, close_fds=True).pid)
+                    proc = subprocess.Popen(child_cmd, env=env, shell=True, close_fds=True)
+                    children_pids.add(proc.pid)
+                    procs.append(proc)
                 except OSError, e:
                     sys.stderr.write("Error spawning child '%s': %s\n"
                                      % (child_cmd, e))
-        child_reaper.set_children_pids(children_pids)
-    # Check once after the mainloop is running, just in case the exit
-    # conditions are satisfied before we even enter the main loop.
-    # (Programming with unix the signal API sure is annoying.)
-    def check_once():
-        child_reaper.check()
-        return False # Only call once
-    gobject.timeout_add(0, check_once)
 
     _cleanups.insert(0, app.cleanup)
-    if app.run():
+    signal.signal(signal.SIGTERM, app.signal_quit)
+    signal.signal(signal.SIGINT, app.signal_quit)
+    try:
+        e = app.run()
+    except KeyboardInterrupt:
+        e = 0
+    if e>0:
         log.info("upgrading: not cleaning up Xvfb or socket")
         # Upgrading, so leave X server running
         # and don't delete the new socket (not ours)
