@@ -1,5 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
+# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -17,4 +18,4 @@ def add_client_options(parser):
     add_notray_option(parser, ", this will also disable notifications!")
 
 def get_machine_id():
-    return  ""
+    return  u""
