@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 # This file is part of Parti.
-# Copyright (C) 2011, 2012 Antoine Martin <antoine@nagafix.co.uk>
+# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -22,10 +22,15 @@ class Win32Tray:
 		self.balloon_click_callback = None
 
 	def getHWND(self):
+		if self.tray_widget is None:
+			return	None
 		return	self.tray_widget.hwnd
 
 	def close(self):
-		self.tray_widget.close()
+		log("close() tray_widget=%s", self.tray_widget)
+		if self.tray_widget:
+			self.tray_widget.close()
+			self.tray_widget = None
 
 
 	#****************************************************************
@@ -74,7 +79,7 @@ class Win32Tray:
 							self.balloon_click_callback()
 							self.balloon_click_callback = None
 				else:
-					log.debug("unkown message")
+					log.debug("unknown win32 message: %s / %s / %s", msg, wParam, lParam)
 				# Pass all messages to the original WndProc
 				try:
 					return win32gui.CallWindowProc(self.old_win32_proc, hWnd, msg, wParam, lParam)
