@@ -1,6 +1,6 @@
 # This file is part of Parti.
 # Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
-# Copyright (C) 2011, 2012 Antoine Martin <antoine@nagafix.co.uk>
+# Copyright (C) 2011-2013 Antoine Martin <antoine@devloop.org.uk>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
@@ -9,7 +9,7 @@
 
 import sys
 import os
-from wimpiggy.gobject_compat import import_gtk, import_gdk
+from wimpiggy.gobject_compat import import_gtk, import_gdk, is_gtk3
 gtk = import_gtk()
 gdk = import_gdk()
 _display = gdk.get_display()
@@ -25,16 +25,18 @@ log = Logger()
 class ClientExtras(ClientExtrasBase):
     def __init__(self, client, opts, conn):
         ClientExtrasBase.__init__(self, client, opts, conn)
-        try:
-            from xpra.platform.gdk_clipboard import GDKClipboardProtocolHelper
-            self.setup_clipboard_helper(GDKClipboardProtocolHelper)
-        except ImportError, e:
-            log.error("GDK Clipboard failed to load: %s - using 'Default Clipboard' fallback", e)
-            self.setup_clipboard_helper(DefaultClipboardProtocolHelper)
+        if not is_gtk3():       #currently broken with gtk3
+            try:
+                from xpra.platform.gdk_clipboard import GDKClipboardProtocolHelper
+                self.setup_clipboard_helper(GDKClipboardProtocolHelper)
+            except ImportError, e:
+                log.error("GDK Clipboard failed to load: %s - using 'Default Clipboard' fallback", e)
+                self.setup_clipboard_helper(DefaultClipboardProtocolHelper)
         self.setup_menu(True)
         self.setup_tray(opts.no_tray, opts.delay_tray, opts.tray_icon)
-        self.setup_xprops(opts.pulseaudio)
+        self.setup_xprops()
         self.setup_x11_bell()
+        self.setup_pa_audio_tagging()
         self.has_dbusnotify = False
         self.has_pynotify = False
         if opts.notifications:
@@ -47,6 +49,13 @@ class ClientExtras(ClientExtrasBase):
             self.hide_tray()
             self.tray_widget = None
 
+    def setup_pa_audio_tagging(self):
+        try:
+            from xpra.sound.pulseaudio_util import add_audio_tagging_env
+            add_audio_tagging_env(self.get_tray_icon_filename(None))
+        except Exception, e:
+            log("failed to set pulseaudio audio tagging: %s", e)
+
     def get_data_dir(self):
         #is there a better/cleaner way?
         options = ["/usr/share/xpra", "/usr/local/share/xpra"]
@@ -178,7 +187,7 @@ class ClientExtras(ClientExtrasBase):
         if self._is_ubuntu_11_10_or_later() and self.setup_appindicator(delay_tray, tray_icon_filename):
             return
         if not self.setup_statusicon(delay_tray, tray_icon_filename):
-            log.error("failed to setup system-tray")
+            log.error("failed to setup tray icon")
 
     def setup_dbusnotify(self):
         self.dbus_id = os.environ.get("DBUS_SESSION_BUS_ADDRESS", "")
@@ -215,14 +224,10 @@ class ClientExtras(ClientExtrasBase):
         except ImportError, e:
             log.error("cannot import x11 bell bindings (will use gtk fallback) : %s", e)
 
-    def setup_xprops(self, pulseaudio):
+    def setup_xprops(self):
         self.ROOT_PROPS = {
             "RESOURCE_MANAGER": "resource-manager"
             }
-        if pulseaudio:
-            self.ROOT_PROPS["PULSE_COOKIE"] = "pulse-cookie"
-            self.ROOT_PROPS["PULSE_ID"] = "pulse-id"
-            self.ROOT_PROPS["PULSE_SERVER"] = "pulse-server"
         def setup_xprop_xsettings(client):
             log.debug("setup_xprop_xsettings(%s)", client)
             try:
@@ -239,23 +244,27 @@ class ClientExtras(ClientExtrasBase):
         self.client.connect("handshake-complete", setup_xprop_xsettings)
 
     def _handle_xsettings_changed(self, *args):
-        blob = self._xsettings_watcher.get_settings_blob()
+        try:
+            blob = self._xsettings_watcher.get_settings_blob()
+        except:
+            log.error("failed to get XSETTINGS", exc_info=True)
+            return
         log("xsettings_changed new value=%s", blob)
         if blob is not None:
-            self.client.send(["server-settings", {"xsettings-blob": blob}])
+            self.client.send("server-settings", {"xsettings-blob": blob})
 
     def _handle_root_prop_changed(self, obj, prop, value):
         log("root_prop_changed: %s=%s", prop, value)
         assert prop in self.ROOT_PROPS
         if value is not None and self.client.xsettings_tuple:
-            self.client.send(["server-settings", {self.ROOT_PROPS[prop]: value.encode("utf-8")}])
+            self.client.send("server-settings", {self.ROOT_PROPS[prop]: value.encode("utf-8")})
 
     def system_bell(self, window, device, percent, pitch, duration, bell_class, bell_id, bell_name):
         if self.has_x11_bell:
             try:
                 from wimpiggy.error import trap, XError
                 from wimpiggy.lowlevel.bindings import device_bell      #@UnresolvedImport
-                trap.call(device_bell, window, device, bell_class, bell_id, percent, bell_name)
+                trap.call_synced(device_bell, window, device, bell_class, bell_id, percent, bell_name)
                 return
             except XError, e:
                 log.error("error using device_bell: %s, will fallback to gdk beep from now on", e)
@@ -321,12 +330,11 @@ class ClientExtras(ClientExtrasBase):
     def exec_get_keyboard_data(self, cmd):
         # Find the client's current keymap so we can send it to the server:
         try:
-            import subprocess
-            process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
-            (out,_) = process.communicate(None)
-            if process.returncode==0:
+            from xpra.scripts.exec_util import safe_exec
+            returncode, out, _ = safe_exec(cmd)
+            if returncode==0:
                 return out.decode('utf-8')
-            log.error("'%s' failed with exit code %s", cmd, process.returncode)
+            log.error("'%s' failed with exit code %s", cmd, returncode)
         except Exception, e:
             log.error("error running '%s': %s", cmd, e)
         return None
